import asyncio
import sys
import json
import re
import typing as ty
from tempfile import gettempdir
from pathlib import Path
from shutil import copyfile
import logging
import attrs
from pydra.engine.job import Job, save
from pydra.workers import base


logger = logging.getLogger("pydra.worker")

if ty.TYPE_CHECKING:
    from pydra.engine.result import Result


@attrs.define
class SlurmWorker(base.Worker):
    """A worker to execute tasks on SLURM systems."""

    _cmd = "sbatch"
    _sacct_re = re.compile(
        "(?P<jobid>\\d*) +(?P<status>\\w*)\\+? +" "(?P<exit_code>\\d+):\\d+"
    )

    poll_delay: int = attrs.field(default=1, converter=base.ensure_non_negative)
    sbatch_args: str = ""
    error: dict[str, ty.Any] = attrs.field(factory=dict)

    def __getstate__(self) -> dict[str, ty.Any]:
        """Return state for pickling."""
        state = super().__getstate__()
        del state["error"]
        return state

    def __setstate__(self, state: dict[str, ty.Any]):
        """Set state for unpickling."""
        state["error"] = {}
        super().__setstate__(state)

    def _prepare_runscripts(self, job, interpreter="/bin/sh", rerun=False):
        if isinstance(job, Job):
            cache_root = job.cache_root
            ind = None
            uid = job.uid
        else:
            assert isinstance(job, tuple), f"Expecting a job or a tuple, not {job!r}"
            assert len(job) == 2, f"Expecting a tuple of length 2, not {job!r}"
            ind = job[0]
            cache_root = job[-1].cache_root
            uid = f"{job[-1].uid}_{ind}"

        script_dir = cache_root / f"{self.plugin_name()}_scripts" / uid
        script_dir.mkdir(parents=True, exist_ok=True)
        if ind is None:
            if not (script_dir / "_job.pklz").exists():
                save(script_dir, job=job)
        else:
            copyfile(job[1], script_dir / "_job.pklz")

        job_pkl = script_dir / "_job.pklz"
        if not job_pkl.exists() or not job_pkl.stat().st_size:
            raise Exception("Missing or empty job!")

        batchscript = script_dir / f"batchscript_{uid}.sh"
        python_string = (
            f"""'from pydra.engine.job import load_and_run; """
            f"""load_and_run("{job_pkl}", rerun={rerun}) '"""
        )
        bcmd = "\n".join(
            (
                f"#!{interpreter}",
                f"#SBATCH --output={script_dir / 'slurm-%j.out'}",
                f"{sys.executable} -c " + python_string,
            )
        )
        with batchscript.open("wt") as fp:
            fp.writelines(bcmd)
        return script_dir, batchscript

    async def run(self, job: "Job[base.TaskType]", rerun: bool = False) -> "Result":
        """Worker submission API."""
        script_dir, batch_script = self._prepare_runscripts(job, rerun=rerun)
        if (script_dir / script_dir.parts[1]) == gettempdir():
            logger.warning("Temporary directories may not be shared across computers")
        script_dir = job.cache_root / f"{self.plugin_name()}_scripts" / job.uid
        sargs = self.sbatch_args.split()
        jobname = re.search(r"(?<=-J )\S+|(?<=--job-name=)\S+", self.sbatch_args)
        if not jobname:
            jobname = ".".join((job.name, job.uid))
            sargs.append(f"--job-name={jobname}")
        output = re.search(r"(?<=-o )\S+|(?<=--output=)\S+", self.sbatch_args)
        if not output:
            output_file = str(script_dir / "slurm-%j.out")
            sargs.append(f"--output={output_file}")
        error = re.search(r"(?<=-e )\S+|(?<=--error=)\S+", self.sbatch_args)
        if not error:
            error_file = str(script_dir / "slurm-%j.err")
            sargs.append(f"--error={error_file}")
        else:
            error_file = None
        sargs.append(str(batch_script))
        # TO CONSIDER: add random sleep to avoid overloading calls
        rc, stdout, stderr = await base.read_and_display_async(
            "sbatch", *sargs, hide_display=True
        )
        jobid = re.search(r"\d+", stdout)
        if rc:
            raise RuntimeError(f"Error returned from sbatch: {stderr}")
        elif not jobid:
            raise RuntimeError("Could not extract job ID")
        jobid = jobid.group()
        if error_file:
            error_file = error_file.replace("%j", jobid)
        self.error[jobid] = error_file.replace("%j", jobid)
        # intermittent polling
        while True:
            # 3 possibilities
            # False: job is still pending/working
            # True: job is complete
            # Exception: Polling / job failure
            done = await self._poll_job(jobid)
            if done:
                if (
                    done in ["CANCELLED", "TIMEOUT", "PREEMPTED"]
                    and "--no-requeue" not in self.sbatch_args
                ):
                    # loading info about job with a specific uid
                    info_file = job.cache_root / f"{job.uid}_info.json"
                    if info_file.exists():
                        checksum = json.loads(info_file.read_text())["checksum"]
                        if (job.cache_root / f"{checksum}.lock").exists():
                            # for pyt3.8 we could you missing_ok=True
                            (job.cache_root / f"{checksum}.lock").unlink()
                    cmd_re = ("scontrol", "requeue", jobid)
                    await base.read_and_display_async(*cmd_re, hide_display=True)
                else:
                    return True
            await asyncio.sleep(self.poll_delay)

    async def _poll_job(self, jobid):
        cmd = ("squeue", "-h", "-j", jobid)
        logger.debug(f"Polling job {jobid}")
        rc, stdout, stderr = await base.read_and_display_async(*cmd, hide_display=True)
        if not stdout or "slurm_load_jobs error" in stderr:
            # job is no longer running - check exit code
            status = await self._verify_exit_code(jobid)
            return status
        return False

    async def _verify_exit_code(self, jobid):
        cmd = ("sacct", "-n", "-X", "-j", jobid, "-o", "JobID,State,ExitCode")
        _, stdout, _ = await base.read_and_display_async(*cmd, hide_display=True)
        if not stdout:
            raise RuntimeError("Job information not found")
        m = self._sacct_re.search(stdout)
        error_file = self.error[jobid]
        if int(m.group("exit_code")) != 0 or m.group("status") != "COMPLETED":
            if m.group("status") in ["CANCELLED", "TIMEOUT", "PREEMPTED"]:
                return m.group("status")
            elif m.group("status") in ["RUNNING", "PENDING"]:
                return False
            # TODO: potential for requeuing
            # parsing the error message
            error_line = Path(error_file).read_text().split("\n")[-2]
            if "Exception" in error_line:
                error_message = error_line.replace("Exception: ", "")
            elif "Error" in error_line:
                error_message = error_line.replace("Exception: ", "")
            else:
                error_message = "Job failed (unknown reason - TODO)"
            raise Exception(error_message)
        return True


# Alias so it can be referred to as slurm.Worker
Worker = SlurmWorker



import itertools
import inspect
from pathlib import Path
import collections.abc
import os
from copy import copy
import sys
import types
import typing as ty
import logging
import attrs
from fileformats import field, core, generic
from pydra.utils.general import add_exc_note
from pydra.utils.mount_identifier import MountIndentifier

try:
    from typing import get_origin, get_args
except ImportError:
    # Python < 3.8
    from typing import get_origin, get_args  # type: ignore

if sys.version_info >= (3, 10):
    UNION_TYPES = (ty.Union, types.UnionType)
else:
    UNION_TYPES = (ty.Union,)

if ty.TYPE_CHECKING:
    from pydra.engine.lazy import LazyField

logger = logging.getLogger("pydra")

NO_GENERIC_ISSUBCLASS = sys.version_info.major == 3 and sys.version_info.minor < 10

if NO_GENERIC_ISSUBCLASS:
    from typing_utils import issubtype

try:
    import numpy
except ImportError:
    HAVE_NUMPY = False
else:
    HAVE_NUMPY = True

T = ty.TypeVar("T")
TypeOrAny = ty.Union[type, ty.Any]


# These are special types that are checked for in the construction of input/output specs
# and special converters inserted into the attrs fields.


class MultiInputObj(list, ty.Generic[T]):
    pass


MultiInputFile = MultiInputObj[generic.File]


# Since we can't create a NewType from a type union, we add a dummy type to the union
# so we can detect the MultiOutput in the input/output task creation
class MultiOutputType:
    pass


MultiOutputObj = ty.Union[list, object, MultiOutputType]
MultiOutputFile = ty.Union[generic.File, ty.List[generic.File], MultiOutputType]

OUTPUT_TEMPLATE_TYPES = (
    Path,
    ty.List[Path],
    ty.Union[Path, bool],
    ty.Union[ty.List[Path], bool],
    ty.List[ty.List[Path]],
)


class StateArray(ty.List[T]):
    """an array of values from, or to be split over in an array of nodes (see TaskBase.split()),
    multiple nodes of the same task. Used in type-checking to differentiate between list
    types and values for multiple nodes
    """

    def __repr__(self):
        return f"{type(self).__name__}(" + ", ".join(repr(i) for i in self) + ")"


class TypeParser(ty.Generic[T]):
    """A callable which can be used as a converter for attrs.fields to check whether an
    object or LazyField matches the specified field type, or can be
    coerced into it (given the criteria passed on initialisation of the checker).
    Nested container type are expanded and each of their type args are checked/coerced
    against corresponding parts of the object.

    Parameters
    ----------
    tp : type
        the type objects will be coerced to
    coercible: Iterable[ty.Tuple[type or Any, type or Any]], optional
        limits coercing between the pairs of types where they appear within the
        tree of more complex nested container types. If None, then all types are
        coercible except explicitly excluded
    not_coercible: Iterable[ty.Tuple[type or Any, type or Any]], optional
        excludes the limits coercing between the pairs of types where they appear within
        the tree of more complex nested container types. Overrides 'coercible' to enable
        you to carve out exceptions, such as TypeParser(list, coercible=[(ty.Iterable, list)],
        not_coercible=[(str, list)])
    superclass_auto_cast : bool
        Allow lazy fields to pass the type check if their types are superclasses of the
        specified pattern (instead of matching or being subclasses of the pattern)
    label : str
        the label to be used to identify the type parser in error messages. Especially
        useful when TypeParser is used as a converter in attrs.fields
    match_any_of_union : bool
        match if any of the options in the union are a subclass (but not necessarily all)
    """

    tp: ty.Type[T]
    coercible: ty.List[ty.Tuple[TypeOrAny, TypeOrAny]]
    not_coercible: ty.List[ty.Tuple[TypeOrAny, TypeOrAny]]
    superclass_auto_cast: bool
    label: str
    match_any_of_union: bool

    COERCIBLE_DEFAULT: ty.Tuple[ty.Tuple[type, type], ...] = (
        (
            (ty.Sequence, ty.Sequence),
            (ty.Sequence, collections.abc.Set),
            (collections.abc.Set, ty.Sequence),
            (ty.Mapping, ty.Mapping),
            (Path, os.PathLike),
            (str, os.PathLike),
            (os.PathLike, Path),
            (os.PathLike, str),
            (ty.Any, MultiInputObj),
            (int, float),
            (field.Integer, float),
            (int, field.Decimal),
        )
        + tuple((f, f.primitive) for f in field.Singular.subclasses() if f.primitive)
        + tuple((f.primitive, f) for f in field.Singular.subclasses() if f.primitive)
    )

    if HAVE_NUMPY:
        COERCIBLE_DEFAULT += (
            (numpy.integer, int),
            (numpy.floating, float),
            (numpy.bool_, bool),
            (numpy.integer, float),
            (numpy.character, str),
            (numpy.complexfloating, complex),
            (numpy.bytes_, bytes),
            (numpy.ndarray, ty.Sequence),
            (ty.Sequence, numpy.ndarray),
        )

    NOT_COERCIBLE_DEFAULT = ((str, ty.Sequence), (ty.Sequence, str))

    def __init__(
        self,
        tp,
        coercible: ty.Optional[
            ty.Iterable[ty.Tuple[TypeOrAny, TypeOrAny]]
        ] = COERCIBLE_DEFAULT,
        not_coercible: ty.Optional[
            ty.Iterable[ty.Tuple[TypeOrAny, TypeOrAny]]
        ] = NOT_COERCIBLE_DEFAULT,
        superclass_auto_cast: bool = False,
        label: str = "",
        match_any_of_union: bool = False,
    ):
        def expand_pattern(t):
            """Recursively expand the type arguments of the target type in nested tuples"""
            if t is inspect._empty:
                return None
            origin = get_origin(t)
            if origin is None:
                return t
            args = get_args(t)
            if not args or args == (Ellipsis,):  # Not sure Ellipsis by itself is valid
                # If no args were provided, or those arguments were an ellipsis
                assert isinstance(origin, type)
                return origin
            if origin not in UNION_TYPES + (type,) and not issubclass(
                origin, ty.Iterable
            ):
                raise TypeError(
                    f"TypeParser doesn't know how to handle args ({args}) for {origin} "
                    f"types{self.label_str}"
                )
            return (origin, [expand_pattern(a) for a in args])

        self.label = label
        self.tp = tp
        self.coercible = (
            list(coercible) if coercible is not None else [(ty.Any, ty.Any)]
        )
        self.not_coercible = list(not_coercible) if not_coercible is not None else []
        self.pattern = expand_pattern(tp)
        self.superclass_auto_cast = superclass_auto_cast
        self.match_any_of_union = match_any_of_union

    def __call__(self, obj: ty.Any) -> T:
        """Attempts to coerce the object to the specified type, unless the value is
        a LazyField where the type of the field is just checked instead or an
        attrs.NOTHING where it is simply returned.

        Parameters
        ----------
        obj : ty.Any
            the object to coerce/check-type

        Returns
        -------
        T
            the coerced object

        Raises
        ------
        TypeError
            if the coercion is not possible, or not specified by the
            `coercible`/`not_coercible` parameters, then a TypeError is raised
        """

        coerced: T
        if obj is attrs.NOTHING:
            coerced = attrs.NOTHING  # type: ignore[assignment]
        elif is_lazy(obj):
            try:
                self.check_type(obj._type)
            except TypeError as e:
                if self.superclass_auto_cast:
                    try:
                        # Check whether the type of the lazy field isn't a superclass of
                        # the type to check against, and if so, allow it due to permissive
                        # typing rules.
                        TypeParser(obj._type, match_any_of_union=True).check_type(
                            self.tp
                        )
                    except TypeError:
                        raise TypeError(
                            f"Incorrect type for lazy field{self.label_str}: {obj._type!r} "
                            f"is not a subclass or superclass of {self.tp} (and will not "
                            "be able to be coerced to one that is)"
                        ) from e
                    else:
                        logger.info(
                            "Connecting lazy field %s to %s%s via permissive typing that "
                            "allows super-to-sub type connections",
                            obj,
                            self.tp,
                            self.label_str,
                        )
                else:
                    raise TypeError(
                        f"Incorrect type for lazy field{self.label_str}: {obj._type!r} "
                        f"is not a subclass of {self.tp} (and will not be able to be "
                        "coerced to one that is)"
                    ) from e
            coerced = obj  # type: ignore
            if obj._type is not ty.Any:
                # Used to check whether the type of the field can be changed
                obj._type_checked = True
        elif isinstance(obj, StateArray):
            coerced = StateArray(self(o) for o in obj)  # type: ignore[assignment]
        else:
            try:
                coerced = self.coerce(obj)
            except TypeError as e:
                if obj is None:
                    raise TypeError(
                        f"Mandatory field{self.label_str} of type {self.tp} was not "
                        "provided a value (i.e. a value that wasn't None) "
                    ) from None
                raise TypeError(
                    f"Incorrect type for field{self.label_str}: {obj!r} is not of type "
                    f"{self.tp} (and cannot be coerced to it)"
                ) from e
        return coerced

    def coerce(self, object_: ty.Any) -> T:
        """Attempts to coerce the given object to the type of the specified type"""
        if self.pattern is None:
            return object_

        def expand_and_coerce(obj, pattern: ty.Union[type, tuple]):
            """Attempt to expand the object along the lines of the coercion pattern"""
            if obj is attrs.NOTHING:
                return attrs.NOTHING
            if not isinstance(pattern, tuple):
                return coerce_basic(obj, pattern)
            origin, pattern_args = pattern
            if origin == MultiInputObj:
                return coerce_multi_input(obj, pattern_args)
            if origin in UNION_TYPES:
                return coerce_union(obj, pattern_args)
            if origin is type:
                return coerce_type(obj, pattern_args)
            if not self.is_instance(obj, origin):
                self.check_coercible(obj, origin)
                type_ = origin
            else:
                type_ = type(obj)
            if issubclass(type_, ty.Mapping):
                return coerce_mapping(obj, type_, pattern_args)
            try:
                obj_args = list(obj)
            except TypeError as e:
                msg = (
                    f" (part of coercion from {object_!r} to {self.pattern}"
                    if obj is not object_
                    else ""
                )
                raise TypeError(
                    f"Could not coerce to {type_} as {obj!r} is not iterable{msg}{self.label_str}"
                ) from e
            if issubclass(origin, tuple):
                return coerce_tuple(type_, obj_args, pattern_args)
            if issubclass(origin, ty.Iterable):
                return coerce_sequence(type_, obj_args, pattern_args)
            assert False, f"Coercion from {obj!r} to {pattern} is not handled"

        def coerce_basic(obj, pattern):
            """Coerce an object to a "basic types" like `int`, `float`, `bool`, `Path`
            and `File` in contrast to compound types like `list[int]`,
            `dict[str, str]` and `dict[str, list[int]]`"""
            if self.is_instance(obj, pattern):
                return obj
            self.check_coercible(obj, pattern)
            return coerce_obj(obj, pattern)

        def coerce_union(obj, pattern_args):
            """Coerce an object into the first type in a Union construct that it is
            coercible into"""
            reasons = []
            for arg in pattern_args:
                try:
                    return expand_and_coerce(obj, arg)
                except TypeError as e:
                    reasons.append(e)
            raise TypeError(
                f"Could not coerce object, {obj!r}, to any of the union types "
                f"{pattern_args}{self.label_str}:\n\n"
                + "\n\n".join(f"{a} -> {e}" for a, e in zip(pattern_args, reasons))
            )

        def coerce_multi_input(obj, pattern_args):
            # Attempt to coerce the object into arg type of the MultiInputObj first,
            # and if that fails, try to coerce it into a list of the arg type
            try:
                return coerce_sequence(list, obj, pattern_args)
            except TypeError as e1:
                try:
                    return [expand_and_coerce(obj, pattern_args[0])]
                except TypeError as e2:
                    raise TypeError(
                        f"Could not coerce object ({obj!r}) to MultiInputObj[{pattern_args[0]}] "
                        f"either as sequence of {pattern_args[0]} ({e1}) or a single {pattern_args[0]} "
                        f"object to be wrapped in a list {e2}"
                    ) from e2

        def coerce_mapping(
            obj: ty.Mapping, type_: ty.Type[ty.Mapping], pattern_args: list
        ):
            """Coerce a mapping (e.g. dict)"""
            key_pattern, val_pattern = pattern_args
            items: ty.Iterable[ty.Tuple[ty.Any, ty.Any]]
            try:
                items = obj.items()
            except AttributeError as e:
                msg = (
                    f" (part of coercion from {object_} to {self.pattern}"
                    if obj is not object_
                    else ""
                )
                raise TypeError(
                    f"Could not coerce to {type_} as {obj} is not a mapping type{msg}{self.label_str}"
                ) from e
            return coerce_obj(
                {
                    expand_and_coerce(k, key_pattern): expand_and_coerce(v, val_pattern)
                    for k, v in items
                },
                type_,
            )

        def coerce_tuple(
            type_: ty.Type[ty.Sequence],
            obj_args: list,
            pattern_args: list,
        ):
            """coerce to a tuple object"""
            if pattern_args[-1] is Ellipsis:
                pattern_args = itertools.chain(  # type: ignore[assignment]
                    pattern_args[:-2], itertools.repeat(pattern_args[-2])
                )
            elif len(pattern_args) != len(obj_args):
                raise TypeError(
                    f"Incorrect number of items in tuple, expected "
                    f"{len(pattern_args)}, got {len(obj_args)}{self.label_str}"
                )
            return coerce_obj(
                [expand_and_coerce(o, p) for o, p in zip(obj_args, pattern_args)], type_
            )

        def coerce_sequence(
            type_: ty.Type[ty.Sequence], obj_args: list, pattern_args: list
        ):
            """Coerce a non-tuple sequence object (e.g. list, ...)"""
            assert len(pattern_args) == 1
            return coerce_obj(
                [expand_and_coerce(o, pattern_args[0]) for o in obj_args], type_
            )

        def coerce_type(type_: ty.Type[ty.Any], pattern_args: ty.List[ty.Type[ty.Any]]):
            if not any(issubclass(type_, t) for t in pattern_args):
                raise TypeError(
                    f"{type_} is not one of the specified types {pattern_args}{self.label_str}"
                )
            return type_

        def coerce_obj(obj, type_):
            """Attempt to do the innermost (i.e. non-nested) coercion and fail with
            helpful message
            """
            try:
                return type_(obj)
            except (TypeError, ValueError) as e:
                msg = (
                    f" (part of coercion from {object_} to {self.pattern}"
                    if obj is not object_
                    else ""
                )
                raise TypeError(
                    f"Cannot coerce {obj!r} into {type_}{msg}{self.label_str}"
                ) from e

        return expand_and_coerce(object_, self.pattern)

    def check_type(self, type_: ty.Type[ty.Any]):
        """Checks the given type to see whether it matches or is a subtype of the
        specified type or whether coercion rule is specified between the types

        Parameters
        ----------
        type_ : ty.Type[ty.Any]
            the type to check whether it is coercible into the specified type

        Raises
        ------
        TypeError
            if the type is not either the specified type, a sub-type or coercible to it
        """
        if self.pattern is None or type_ is ty.Any:
            return
        if self.is_subclass(type_, StateArray):
            args = get_args(type_)
            if not args:
                raise TypeError("Splits without any type arguments are invalid")
            if len(args) > 1:
                raise TypeError(
                    f"Splits with more than one type argument ({args}) are invalid{self.label_str}"
                )
            return self.check_type(args[0])

        def expand_and_check(tp, pattern: ty.Union[type, tuple]):
            """Attempt to expand the object along the lines of the coercion pattern"""
            if not isinstance(pattern, tuple):
                return check_basic(tp, pattern)
            pattern_origin, pattern_args = pattern
            if pattern_origin in UNION_TYPES:
                return check_union(tp, pattern_args)
            tp_origin = get_origin(tp)
            if tp_origin is None:
                if issubclass(tp, pattern_origin):
                    raise TypeError(
                        f"Type {tp} wasn't declared with type args required to match pattern "
                        f"{pattern_args}, when matching {type_} to {self.pattern}"
                    )
                raise TypeError(
                    f"{tp} doesn't match pattern {pattern}, when matching {type_} to "
                    f"{self.pattern}{self.label_str}"
                )
            tp_args = get_args(tp)
            self.check_type_coercible(tp_origin, pattern_origin)
            if issubclass(pattern_origin, ty.Mapping):
                return check_mapping(tp_args, pattern_args)
            if issubclass(pattern_origin, tuple):
                if not issubclass(tp_origin, tuple):
                    assert len(tp_args) == 1
                    tp_args += (Ellipsis,)
                return check_tuple(tp_args, pattern_args)
            return check_sequence(tp_args, pattern_args)

        def check_basic(tp, target):
            # Note that we are deliberately more permissive than typical type-checking
            # here, allowing parents of the target type as well as children,
            # to avoid users having to cast from loosely typed tasks to strict ones
            if self.match_any_of_union and get_origin(tp) is ty.Union:
                reasons = []
                tp_args = get_args(tp)
                for tp_arg in tp_args:
                    if self.is_subclass(tp_arg, target):
                        return
                    try:
                        self.check_coercible(tp_arg, target)
                    except TypeError as e:
                        reasons.append(e)
                    else:
                        return
                if reasons:
                    raise TypeError(
                        f"Cannot coerce any union args {tp_arg} to {target}"
                        f"{self.label_str}:\n\n"
                        + "\n\n".join(f"{a} -> {e}" for a, e in zip(tp_args, reasons))
                    )
            if not self.is_subclass(tp, target):
                self.check_type_coercible(tp, target)

        def check_union(tp, pattern_args):
            if get_origin(tp) in UNION_TYPES:
                tp_args = get_args(tp)
                for tp_arg in tp_args:
                    reasons = []
                    for pattern_arg in pattern_args:
                        try:
                            expand_and_check(tp_arg, pattern_arg)
                        except TypeError as e:
                            reasons.append(e)
                        else:
                            reasons = []
                            break
                    if self.match_any_of_union and len(reasons) < len(tp_args):
                        # Just need one of the union args to match
                        return
                    if reasons:
                        determiner = "any" if self.match_any_of_union else "all"
                        raise TypeError(
                            f"Cannot coerce {tp} to ty.Union["
                            f"{', '.join(str(a) for a in pattern_args)}]{self.label_str}, "
                            f"because {tp_arg} cannot be coerced to {determiner} of its args:\n\n"
                            + "\n\n".join(
                                f"{a} -> {e}" for a, e in zip(pattern_args, reasons)
                            )
                        )
                return
            reasons = []
            for pattern_arg in pattern_args:
                try:
                    return expand_and_check(tp, pattern_arg)
                except TypeError as e:
                    reasons.append(e)
            raise TypeError(
                f"Cannot coerce {tp} to any of the union types:\n\n"
                + "\n\n".join(f"{a} -> {e}" for a, e in zip(pattern_args, reasons))
            )

        def check_mapping(tp_args, pattern_args):
            key_pattern, val_pattern = pattern_args
            key_tp, val_tp = tp_args
            expand_and_check(key_tp, key_pattern)
            expand_and_check(val_tp, val_pattern)

        def check_tuple(tp_args, pattern_args):
            if pattern_args[-1] is Ellipsis:
                if len(pattern_args) == 1:  # matches anything
                    return
                if tp_args[-1] is Ellipsis:
                    return expand_and_check(tp_args[0], pattern_args[0])
                for arg in tp_args:
                    expand_and_check(arg, pattern_args[0])
                return
            elif tp_args[-1] is Ellipsis:
                for pattern_arg in pattern_args:
                    expand_and_check(tp_args[0], pattern_arg)
                return
            if len(tp_args) != len(pattern_args):
                raise TypeError(
                    f"Wrong number of type arguments in tuple {tp_args}  compared to pattern "
                    f"{pattern_args} in attempting to match {type_} to {self.pattern}{self.label_str}"
                )
            for t, p in zip(tp_args, pattern_args):
                expand_and_check(t, p)

        def check_sequence(tp_args, pattern_args):
            assert len(pattern_args) == 1
            if tp_args[-1] is Ellipsis:
                tp_args = tp_args[:-1]
                if not tp_args:
                    raise TypeError(
                        "Generic ellipsis type arguments not specific enough to match "
                        f"{pattern_args} in attempting to match {type_} to "
                        f"{self.pattern}{self.label_str}"
                    )
            for arg in tp_args:
                expand_and_check(arg, pattern_args[0])

        try:
            return expand_and_check(type_, self.pattern)
        except TypeError as e:
            # Special handling for MultiInputObjects (which are annoying)
            if not isinstance(self.pattern, tuple) or self.pattern[0] != MultiInputObj:
                raise e
            # Attempt to coerce the object into arg type of the MultiInputObj first,
            # and if that fails, try to coerce it into a list of the arg type
            inner_type_parser = copy(self)
            inner_type_parser.pattern = self.pattern[1][0]
            try:
                inner_type_parser.check_type(type_)
            except TypeError:
                add_exc_note(
                    e,
                    "Also failed to coerce to the arg-type of the MultiInputObj "
                    f"({self.pattern[1][0]})",
                )
                raise e

    def check_coercible(self, source: ty.Any, target: ty.Union[type, ty.Any]):
        """Checks whether the source object is coercible to the target type given the coercion
        rules defined in the `coercible` and `not_coercible` attrs

        Parameters
        ----------
        source : object
            the object to be coerced
        target : type or typing.Any
            the target type for the object to be coerced to

        Raises
        ------
        TypeError
            If the object cannot be coerced into the target type depending on the explicit
            inclusions and exclusions set in the `coercible` and `not_coercible` member attrs
        """
        if (
            isinstance(source, ty.Sequence)
            and issubclass(target, core.FileSet)
            and all(isinstance(p, os.PathLike) for p in source)
        ):
            return True
        self.check_type_coercible(type(source), target, source_repr=repr(source))

    def check_type_coercible(
        self,
        source: ty.Union[type, ty.Any],
        target: ty.Union[type, ty.Any],
        source_repr: ty.Optional[str] = None,
    ):
        """Checks whether the source type is coercible to the target type
        given the coercion rules defined in the `coercible` and `not_coercible` attrs

        Parameters
        ----------
        source : type or typing.Any
            source type to be coerced
        target : type or typing.Any
            target type for the source to be coerced to

        Raises
        ------
        TypeError
            If the source type cannot be coerced into the target type depending on the
            explicit inclusions and exclusions set in the `coercible` and `not_coercible`
            member attrs
        """
        if source_repr is None:
            source_repr = repr(source)
        # Short-circuit the basic cases where the source and target are the same
        if source is target:
            return
        if self.superclass_auto_cast and self.is_subclass(target, source):
            logger.info(
                "Attempting to coerce %s into %s due to super-to-sub class coercion "
                "being permitted",
                source,
                target,
            )
            return
        source_origin = get_origin(source)
        if source_origin is not None:
            source = source_origin

        def matches_criteria(criteria):
            return [
                (src, tgt)
                for src, tgt in criteria
                if self.is_subclass(source, src) and self.is_subclass(target, tgt)
            ]

        def type_name(t):
            try:
                return t.__name__
            except AttributeError:
                return t._name  # typing generics for Python < 3.10

        if not matches_criteria(self.coercible):
            raise TypeError(
                f"Cannot coerce {source_repr} into {target}{self.label_str} as the "
                "coercion doesn't match any of the explicit inclusion criteria: "
                + ", ".join(
                    f"{type_name(s)} -> {type_name(t)}" for s, t in self.coercible
                )
            )
        matches_not_coercible = matches_criteria(self.not_coercible)
        if matches_not_coercible:
            raise TypeError(
                f"Cannot coerce {source_repr} into {target}{self.label_str} as it is explicitly "
                "excluded by the following coercion criteria: "
                + ", ".join(
                    f"{type_name(s)} -> {type_name(t)}"
                    for s, t in matches_not_coercible
                )
            )

    @classmethod
    def matches(cls, obj: ty.Type[ty.Any], target: ty.Type[ty.Any], **kwargs) -> bool:
        """Returns true if the provided type matches the pattern of the TypeParser

        Parameters
        ----------
        type_ : type
            the type to check
        target : type
            the target type to check against
        **kwargs : dict[str, Any], optional
            passed on to TypeParser.__init__

        Returns
        -------
        matches : bool
            whether the type matches the target type factoring in sub-classes and coercible
            pairs
        """
        parser = cls(target, **kwargs)
        try:
            parser.coerce(obj)
        except TypeError:
            return False
        return True

    @classmethod
    def matches_type(
        cls, type_: ty.Type[ty.Any], target: ty.Type[ty.Any], **kwargs
    ) -> bool:
        """Returns true if the provided type matches the pattern of the TypeParser

        Parameters
        ----------
        type_ : type
            the type to check
        target : type
            the target type to check against
        **kwargs : dict[str, Any], optional
            passed on to TypeParser.__init__

        Returns
        -------
        matches : bool
            whether the type matches the target type factoring in sub-classes and coercible
            pairs
        """
        parser = cls(target, **kwargs)
        try:
            parser.check_type(type_)
        except TypeError:
            return False
        return True

    @classmethod
    def is_instance(
        cls,
        obj: object,
        candidates: ty.Union[ty.Type[ty.Any], ty.Sequence[ty.Type[ty.Any]], None],
    ) -> bool:
        """Checks whether the object is an instance of cls or that cls is typing.Any,
        extending the built-in isinstance to check nested type args

        Parameters
        ----------
        obj: object
            the object to check whether it is an instance of one of the candidates
        candidates : type or ty.Iterable[type]
            the candidate types to check the object against
        """
        if candidates is None:
            candidates = [type(None)]
        if not isinstance(candidates, ty.Sequence):
            candidates = [candidates]
        for candidate in candidates:
            if candidate is ty.Any:
                return True
            # Handle ty.Type[*] candidates
            if ty.get_origin(candidate) is type:
                return inspect.isclass(obj) and cls.is_subclass(
                    obj, ty.get_args(candidate)[0]
                )
            if NO_GENERIC_ISSUBCLASS:
                if inspect.isclass(obj):
                    return candidate is type
                if issubtype(type(obj), candidate) or (
                    type(obj) is dict and candidate is ty.Mapping  # noqa: E721
                ):
                    return True
            else:
                if isinstance(obj, candidate):
                    return True
        return False

    @classmethod
    def is_subclass(
        cls,
        klass: ty.Type[ty.Any],
        candidates: ty.Union[ty.Type[ty.Any], ty.Sequence[ty.Type[ty.Any]]],
        any_ok: bool = False,
    ) -> bool:
        """Checks whether the class a is either the same as b, a subclass of b or b is
        typing.Any, extending built-in issubclass to check nested type args

        Parameters
        ----------
        klass : type
            the klass to check whether it is a subclass of one of the candidates
        candidates : type or ty.Iterable[type]
            the candidate types to check the object against
        any_ok : bool
            whether klass=typing.Any should return True or False
        """
        if klass is None:
            # Implicitly convert None to NoneType, like in other typing
            klass = type(None)
        if not isinstance(candidates, ty.Sequence):
            candidates = [candidates]
        if ty.Any in candidates:
            return True
        if klass is ty.Any:
            return any_ok

        origin = get_origin(klass)
        args = get_args(klass)

        for candidate in candidates:
            if candidate is None:
                candidate = type(None)
            candidate_origin = get_origin(candidate)
            candidate_args = get_args(candidate)
            # Handle ty.Type[*] types in klass and candidates
            if origin is type and (candidate is type or candidate_origin is type):
                if candidate is type:
                    return True
                return cls.is_subclass(args[0], candidate_args[0])
            elif origin is type or candidate_origin is type:
                return False
            if NO_GENERIC_ISSUBCLASS:
                if klass is type and candidate is not type:
                    return False
                if issubtype(klass, candidate) or (
                    klass is dict and candidate is ty.Mapping
                ):
                    return True
            else:
                if origin in UNION_TYPES:
                    union_args = (
                        candidate_args
                        if candidate_origin in UNION_TYPES
                        else (candidate,)
                    )
                    matches = all(
                        any(cls.is_subclass(a, c) for c in union_args) for a in args
                    )
                    if matches:
                        return True
                else:
                    if candidate_args and candidate_origin not in UNION_TYPES:
                        if (
                            origin
                            and issubclass(origin, candidate_origin)  # type: ignore[arg-type]
                            and len(args) == len(candidate_args)
                            and all(
                                issubclass(a, c) for a, c in zip(args, candidate_args)
                            )
                        ):
                            return True
                    else:
                        if issubclass(origin if origin else klass, candidate):
                            return True
        return False

    @classmethod
    def contains_type(cls, target: ty.Type[ty.Any], type_: ty.Type[ty.Any]):
        """Checks a potentially nested type for sub-classes of the target type

        Parameters
        ----------
        target : type
            the target type to check for sub-classes of
        type_: type
            the type to check for nested types that are sub-classes of target
        """
        if cls.is_subclass(type_, target):
            return True
        if type_ in (str, bytes, int, bool, float):  # shortcut primitive types
            return False
        type_args = get_args(type_)
        if not type_args:
            return False
        type_origin = get_origin(type_)
        if type_origin in UNION_TYPES:
            for type_arg in type_args:
                if cls.contains_type(target, type_arg):
                    return True
            return False
        if cls.is_subclass(type_origin, ty.Mapping):
            type_key, type_val = type_args
            return cls.contains_type(target, type_key) or cls.contains_type(
                target, type_val
            )
        if cls.is_subclass(type_, (ty.Sequence, MultiOutputObj)):
            if type_args[-1] == Ellipsis:
                type_args = type_args[:-1]
            return any(cls.contains_type(target, a) for a in type_args)
        return False

    @classmethod
    def apply_to_instances(
        cls,
        target_type: ty.Type[ty.Any],
        func: ty.Callable,
        value: ty.Any,
        cache: ty.Optional[ty.Dict[int, ty.Any]] = None,
    ) -> ty.Any:
        """Applies a function to all instances of the given type that are potentially
        nested within the given value, caching previously computed modifications to
        handle repeated elements

        Parameters
        ----------
        target_type : type
            the target type to apply the function to
        func : callable
            the callable object (e.g. function) to apply to the instances
        value : Any
            the value to copy files from (if required)
        cache: dict, optional
            guards against multiple references to the same objects by keeping a cache of
            the modified
        """
        if (
            not cls.is_instance(value, (target_type, ty.Mapping, ty.Sequence))
            or target_type is not str
            and cls.is_instance(value, str)
        ):
            return value
        if cache is None:
            cache = {}
        obj_id = id(value)
        try:
            return cache[obj_id]
        except KeyError:
            pass
        if cls.is_instance(value, target_type):
            modified = func(value)
        elif cls.is_instance(value, ty.Mapping):
            modified = type(value)(  # type: ignore
                (
                    cls.apply_to_instances(target_type, func, key),
                    cls.apply_to_instances(target_type, func, val),
                )
                for (key, val) in value.items()
            )
        else:
            assert cls.is_instance(value, (ty.Sequence, MultiOutputObj))
            args = [cls.apply_to_instances(target_type, func, val) for val in value]
            modified = type(value)(args)  # type: ignore
        cache[obj_id] = modified
        return modified

    @classmethod
    def get_item_type(
        cls, sequence_type: ty.Type[ty.Sequence[T]]
    ) -> ty.Union[ty.Type[T], ty.Any]:
        """Return the type of the types of items in a sequence type

        Parameters
        ----------
        sequence_type: type[Sequence]
            the type to find the type of the items of

        Returns
        -------
        item_type: type or None
            the type of the items
        """
        if not TypeParser.is_subclass(sequence_type, ty.Sequence):
            raise TypeError(
                f"Cannot get item type from {sequence_type}, as it is not a sequence type"
            )
        args = get_args(sequence_type)
        if not args:
            return ty.Any
        if len(args) > 1 and not (len(args) == 2 and args[-1] == Ellipsis):
            raise TypeError(
                f"Cannot get item type from {sequence_type}, as it has multiple "
                f"item types: {args}"
            )
        return args[0]

    @classmethod
    def strip_splits(cls, type_: ty.Type[ty.Any]) -> ty.Tuple[ty.Type, int]:
        """Strips any StateArray types from the outside of the specified type and returns
        the stripped type and the depth it was found at

        Parameters
        ----------
        type_ : ty.Type[ty.Any]
            the type to list the nested sequences of

        Returns
        -------
        inner_type : type
            the inner type once all outer sequences are stripped
        depth : int
            the number of splits outside the inner_type
        """
        depth = 0
        while cls.is_subclass(type_, StateArray) and not cls.is_subclass(type_, str):
            origin = get_origin(type_)
            # If type is a union, pick the first sequence type in the union
            if origin in UNION_TYPES:
                for tp in get_args(type_):
                    if cls.is_subclass(tp, ty.Sequence):
                        type_ = tp
                        break
            type_ = cls.get_item_type(type_)
            depth += 1
        return type_, depth

    @property
    def label_str(self):
        return f" in {self.label} " if self.label else ""

    get_origin = staticmethod(get_origin)
    get_args = staticmethod(get_args)


def is_union(type_: type, args: list[type] = None) -> bool:
    """Checks whether a type is a Union, in either ty.Union[T, U] or T | U form

    Parameters
    ----------
    type_ : type
        the type to check
    args : list[type], optional
        required arguments of the union to check, by default (None) any args will match

    Returns
    -------
    is_union : bool
        whether the type is a Union type
    """
    if ty.get_origin(type_) in UNION_TYPES:
        if args is not None:
            return ty.get_args(type_) == args
        return True
    return False


def is_optional(type_: type) -> bool:
    """Check if the type is Optional, i.e. a union containing None"""
    if is_union(type_):
        return any(a is type(None) or is_optional(a) for a in ty.get_args(type_))
    return False


def is_container(type_: type) -> bool:
    """Check if the type is a container, i.e. a list, tuple, or MultiOutputObj"""
    origin = ty.get_origin(type_)
    if origin is ty.Union:
        return all(is_container(a) for a in ty.get_args(type_))
    tp = origin if origin else type_
    return inspect.isclass(tp) and issubclass(tp, ty.Container)


def is_truthy_falsy(type_: type) -> bool:
    """Check if the type is a truthy type, i.e. not None, bool, or typing.Any"""
    return (
        type_ in (ty.Any, bool, int, str)
        or is_optional(type_)
        or is_container(type_)
        or hasattr(type_, "__bool__")
        or hasattr(type_, "__len__")
    )


def optional_type(type_: type) -> type:
    """Gets the non-None args of an optional type (i.e. a union with a None arg)"""
    if is_optional(type_):
        non_none = [a for a in ty.get_args(type_) if a is not type(None)]
        if len(non_none) == 1:
            return non_none[0]
        return ty.Union[tuple(non_none)]
    return type_


def is_multi_input(type_: type) -> bool:
    """Check if the type is a MultiInputObj"""
    type_ = optional_type(type_)
    return MultiInputObj in (type_, ty.get_origin(type_))


def is_fileset_or_union(type_: type, allow_none: bool | None = None) -> bool:
    """Check if the type is a FileSet or a Union containing a FileSet

    Parameters
    ----------
    type_ : type
        the type to check
    allow_none : bool, optional
        whether to allow None as a valid type, by default None. If None, then None
        is not allowed at the outer layer, but is allowed within a Union

    Returns
    -------
    is_fileset : bool
        whether the type is a FileSet or a Union containing a FileSet
    """
    return is_subclass_or_union(type_, core.FileSet, allow_none=allow_none)


def is_subclass_or_union(
    type_: type, reference: type, allow_none: bool | None = None
) -> bool:
    """Check if the type is a subclass of given reference or a Union containing
    that reference type

    Parameters
    ----------
    type_ : type
        the type to check
    reference : type
        the reference type to check whether the type is a sub-class of or not
    allow_none : bool, optional
        whether to allow None as a valid type, by default None. If None, then None
        is not allowed at the outer layer, but is allowed within a Union

    Returns
    -------
    bool
        whether the type is a FileSet or a Union containing a FileSet
    """
    if type_ is None and allow_none:
        return True
    if is_union(type_):
        return any(
            is_subclass_or_union(
                t, reference, allow_none=allow_none or allow_none is None
            )
            for t in ty.get_args(type_)
        )
    elif not inspect.isclass(type_):
        return False
    return issubclass(type_, reference)


def is_type(*args: ty.Any) -> bool:
    """check that the value is a type or generic"""
    if len(args) == 3:  # attrs validator
        val = args[2]
    elif len(args) != 1:
        raise TypeError(f"is_type() takes 1 or 3 arguments, not {args}")
    else:
        val = args[0]
    return inspect.isclass(val) or ty.get_origin(val)


T = ty.TypeVar("T")
U = ty.TypeVar("U")


def state_array_support(
    function: ty.Callable[T, U],
) -> ty.Callable[T | StateArray[T], U | StateArray[U]]:
    """
    Decorator to convert a allow a function to accept and return StateArray objects,
    where the function is applied to each element of the StateArray.
    """

    def state_array_wrapper(
        value: "T | StateArray[T] | LazyField[T]",
    ) -> "U | StateArray[U] | LazyField[U]":
        if is_lazy(value):
            return value
        if isinstance(value, StateArray):
            return StateArray(function(v) for v in value)
        return function(value)

    return state_array_wrapper


def is_lazy(obj):
    """Check whether an object is a lazy field or has any attribute that is a Lazy Field"""
    from pydra.engine.lazy import LazyField

    return isinstance(obj, LazyField)


def copy_nested_files(
    value: ty.Any,
    dest_dir: os.PathLike,
    supported_modes: generic.FileSet.CopyMode = generic.FileSet.CopyMode.any,
    clashes_to_avoid: set[Path] | None = None,
    **kwargs,
) -> ty.Any:
    """Copies all "file-sets" found within the nested value (e.g. dict, list,...) into the
    destination directory. If no nested file-sets are found then the original value is
    returned. Note that multiple nested file-sets (e.g. a list) will to have unique names
    names (i.e. not differentiated by parent directories) otherwise there will be a path
    clash in the destination directory.

    Parameters
    ----------
    value : Any
        the value to copy files from (if required)
    dest_dir : os.PathLike
        the destination directory to copy the files to
    **kwargs
        passed directly onto FileSet.copy()
    """
    from pydra.utils.typing import TypeParser  # noqa

    cache: ty.Dict[generic.FileSet, generic.FileSet] = {}

    # Set to keep track of file paths that have already been copied
    # to allow FileSet.copy to avoid name clashes
    if clashes_to_avoid is None:
        clashes_to_avoid = set()

    def copy_fileset(fileset: generic.FileSet):
        try:
            return cache[fileset]
        except KeyError:
            pass
        supported = supported_modes
        if any(MountIndentifier.on_cifs(p) for p in fileset.fspaths):
            supported -= generic.FileSet.CopyMode.symlink
        if not all(
            MountIndentifier.on_same_mount(p, dest_dir) for p in fileset.fspaths
        ):
            supported -= generic.FileSet.CopyMode.hardlink
        cp_kwargs = {}

        cp_kwargs.update(kwargs)
        copied = fileset.copy(
            dest_dir=dest_dir,
            supported_modes=supported,
            avoid_clashes=clashes_to_avoid,  # this prevents fname clashes between filesets
            **kwargs,
        )
        cache[fileset] = copied
        return copied

    return TypeParser.apply_to_instances(generic.FileSet, copy_fileset, value)



from .general import (
    get_fields,
    asdict,
    plot_workflow,
    show_workflow,
    task_help,
    print_help,
    unstructure,
    structure,
)
from ._version import __version__

__all__ = [
    "__version__",
    "get_fields",
    "asdict",
    "plot_workflow",
    "show_workflow",
    "task_help",
    "print_help",
    "unstructure",
    "structure",
]



import typing as ty
import inspect
from typing import dataclass_transform
import attrs
from pydra.utils.general import get_fields, asdict
from pydra.compose import base
from pydra.compose.base import (
    ensure_field_objects,
    build_task_class,
    parse_doc_string,
    extract_function_inputs_and_outputs,
    check_explicit_fields_are_none,
    extract_fields_from_class,
)

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job

__all__ = ["arg", "out", "define", "Task", "Outputs"]


@attrs.define
class arg(base.Arg):
    """Argument of a Python task

    Parameters
    ----------
    help: str
        A short description of the input field.
    default : Any, optional
        the default value for the argument
    allowed_values: list, optional
        List of allowed values for the field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    type: type, optional
        The type of the field, by default it is Any
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    """

    pass


@attrs.define
class out(base.Out):
    """Output of a Python task

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    type: type, optional
        The type of the field, by default it is Any
    help: str, optional
        A short description of the input field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    converter: callable, optional
        The converter for the field passed through to the attrs.field, by default it is None
    validator: callable | iterable[callable], optional
        The validator(s) for the field passed through to the attrs.field, by default it is None
    position : int
        The position of the output in the output list, allows for tuple unpacking of
        outputs
    """

    pass


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(arg,),
)
def define(
    wrapped: type | ty.Callable | None = None,
    /,
    inputs: list[str | arg] | dict[str, arg | type] | None = None,
    outputs: list[str | out] | dict[str, out | type] | type | None = None,
    bases: ty.Sequence[type] = (),
    outputs_bases: ty.Sequence[type] = (),
    auto_attribs: bool = True,
    name: str | None = None,
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]] = (),
) -> "Task":
    """
    Create an interface for a function or a class.

    Parameters
    ----------
    wrapped : type | callable | None
        The function or class to create an interface for.
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The inputs to the function or class.
    outputs : list[str | base.Out] | dict[str, base.Out | type] | type | None
        The outputs of the function or class.
    auto_attribs : bool
        Whether to use auto_attribs mode when creating the class.
    name: str | None
        The name of the returned class
    xor: Sequence[str | None] | Sequence[Sequence[str | None]], optional
        Names of args that are exclusive mutually exclusive, which must include
        the name of the current field. If this list includes None, then none of the
        fields need to be set.

    Returns
    -------
    Task
        The task class for the Python function
    """

    def make(wrapped: ty.Callable | type) -> Task:
        if inspect.isclass(wrapped):
            klass = wrapped
            function = klass.function
            class_name = klass.__name__
            check_explicit_fields_are_none(klass, inputs, outputs)
            parsed_inputs, parsed_outputs = extract_fields_from_class(
                Task,
                Outputs,
                klass,
                arg,
                out,
                auto_attribs,
                skip_fields=["function"],
            )
        else:
            if not isinstance(wrapped, ty.Callable):
                raise ValueError(
                    f"wrapped must be a class or a function, not {wrapped!r}"
                )
            klass = None
            function = wrapped
            input_helps, output_helps = parse_doc_string(function.__doc__)
            inferred_inputs, inferred_outputs = extract_function_inputs_and_outputs(
                function, arg, inputs, outputs
            )

            class_name = function.__name__ if name is None else name

            parsed_inputs, parsed_outputs = ensure_field_objects(
                arg_type=arg,
                out_type=out,
                inputs=inferred_inputs,
                outputs=inferred_outputs,
                input_helps=input_helps,
                output_helps=output_helps,
            )
        if "function" in parsed_inputs:
            raise ValueError(
                "The argument 'function' is reserved for a field to hold the function "
                "to be wrapped"
            )

        parsed_inputs["function"] = arg(
            name="function", type=ty.Callable, default=function, hash_eq=True
        )

        defn = build_task_class(
            Task,
            Outputs,
            parsed_inputs,
            parsed_outputs,
            name=class_name,
            klass=klass,
            bases=bases,
            outputs_bases=outputs_bases,
            xor=xor,
        )

        return defn

    if wrapped is not None:
        if not isinstance(wrapped, (ty.Callable, type)):
            raise ValueError(f"wrapped must be a class or a callable, not {wrapped!r}")
        return make(wrapped)
    return make


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class PythonOutputs(base.Outputs):

    @classmethod
    def _from_job(cls, job: "Job[PythonTask]") -> ty.Self:
        """Collect the outputs of a job from a combination of the provided inputs,
        the objects in the output directory, and the stdout and stderr of the process.

        Parameters
        ----------
        job : Job[Task]
            The job whose outputs are being collected.
        outputs_dict : dict[str, ty.Any]
            The outputs of the job, as a dictionary

        Returns
        -------
        outputs : Outputs
            The outputs of the job in dataclass
        """
        outputs = super()._from_job(job)
        for name, val in job.return_values.items():
            setattr(outputs, name, val)
        return outputs


PythonOutputsType = ty.TypeVar("PythonOutputsType", bound=PythonOutputs)


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class PythonTask(base.Task[PythonOutputsType]):

    _executor_name = "function"

    def _run(self, job: "Job[PythonTask]", rerun: bool = True) -> None:
        # Prepare the inputs to the function
        inputs = asdict(self)
        del inputs["function"]
        # Run the actual function
        returned = self.function(**inputs)
        # Collect the outputs and save them into the job.return_values dictionary
        return_names = [f.name for f in get_fields(self.Outputs)]
        if returned is None:
            job.return_values = {nm: None for nm in return_names}
        elif not return_names:
            raise ValueError(
                f"No output fields were specified, but the function returned {returned}"
            )
        elif len(return_names) == 1:
            # if only one element in the fields, everything should be returned together
            job.return_values[return_names[0]] = returned
        elif isinstance(returned, tuple) and len(return_names) == len(returned):
            job.return_values.update(zip(return_names, returned))
        elif isinstance(returned, dict):
            job.return_values.update(
                {key: returned[key] for key in return_names if key in returned}
            )
        else:
            raise RuntimeError(
                f"expected {len(return_names)} elements, but {returned} were returned"
            )


# Alias ShellTask to Task so we can refer to it by shell.Task
Task = PythonTask
Outputs = PythonOutputs



import typing as ty
from copy import deepcopy
from enum import Enum
import attrs
from pydra.engine import lazy
from pydra.utils.general import attrs_values, asdict
from pydra.utils.typing import is_lazy
from pydra.engine.state import State, add_name_splitter, add_name_combiner

if ty.TYPE_CHECKING:
    from pydra.engine.workflow import Workflow
    from pydra.environments.base import Environment
    from pydra.compose import base
    from pydra.engine.hooks import TaskHooks


OutputType = ty.TypeVar("OutputType", bound="base.Outputs")
Splitter = ty.Union[str, ty.Tuple[str, ...]]

_not_set = Enum("_not_set", "NOT_SET")

NOT_SET = _not_set.NOT_SET


@attrs.define
class Node(ty.Generic[OutputType]):
    """A node in a workflow

    Parameters
    ----------
    name : str
        The name of the node
    inputs : Task
        The task of the node
    """

    name: str
    _task: "base.Task[OutputType]"
    _environment: "Environment | None" = None
    _hooks: "TaskHooks | None" = None
    _workflow: "Workflow" = attrs.field(default=None, eq=False, hash=False, repr=False)
    _lzout: OutputType | None = attrs.field(
        init=False, default=None, eq=False, hash=False, repr=False
    )
    _state: State | None = attrs.field(init=False, default=NOT_SET)

    def __attrs_post_init__(self):
        self._set_state()

    class Inputs:
        """A class to wrap the inputs of a node and control access to them so lazy fields
        that will change the downstream state (i.e. with new splits) aren't set after
        the node has been split, combined or its outputs accessed.
        """

        _node: "Node"

        def __init__(self, node: "Node") -> None:
            super().__setattr__("_node", node)

        def __getattr__(self, name: str) -> ty.Any:
            return getattr(self._node._task, name)

        def __getstate__(self) -> ty.Dict[str, ty.Any]:
            return {"_node": self._node}

        def __setstate__(self, state: ty.Dict[str, ty.Any]) -> None:
            super().__setattr__("_node", state["_node"])

        def __setattr__(self, name: str, value: ty.Any) -> None:
            setattr(self._node._task, name, value)
            if is_lazy(value):
                upstream_states = self._node._get_upstream_states()
                if (
                    not self._node._state
                    or self._node._state.other_states != upstream_states
                ):
                    self._node._check_if_outputs_have_been_used(
                        f"cannot set {name!r} input to {value} because it changes the "
                        f"state"
                    )
                    self._node._set_state()

    @property
    def inputs(self) -> Inputs:
        return self.Inputs(self)

    @property
    def input_names(self) -> list[str]:
        return list(attrs_values(self._task).keys())

    @property
    def state(self):
        """Initialise the state of the node just after it has been created (i.e. before
        it has been split or combined) based on the upstream connections
        """
        return self._state

    @property
    def input_values(self) -> tuple[tuple[str, ty.Any]]:
        return tuple(attrs_values(self._task).items())

    @property
    def state_values(self) -> dict[str, ty.Any]:
        """Get the values of the task, scoped by the name of the node to be
        used in the state

        Returns
        -------
        dict[str, Any]
            The values of the task
        """
        return {f"{self.name}.{n}": v for n, v in attrs_values(self._task).items()}

    @property
    def lzout(self) -> OutputType:
        from pydra.utils.general import get_fields

        """The output task of the node populated with lazy fields"""
        if self._lzout is not None:
            return self._lzout
        lazy_fields = {}
        for field in get_fields(self.inputs.Outputs):
            lazy_fields[field.name] = lazy.LazyOutField(
                node=self,
                field=field.name,
                type=field.type,
            )
        outputs = self.inputs.Outputs(**lazy_fields)

        outpt: lazy.LazyOutField
        for outpt in attrs_values(outputs).values():
            # Assign the current node to the lazy fields so they can access the state
            outpt._node = self
            # If the node has a non-empty state, wrap the type of the lazy field in
            # a combination of an optional list and a number of nested StateArrays
            # types based on the number of states the node is split over and whether
            # it has a combiner
            if self._state:
                outpt._type = self._state.nest_output_type(outpt._type)
            # Flag the output lazy fields as being not typed checked (i.e. assigned to
            # another node's inputs) yet. This is used to prevent the user from changing
            # the type of the output after it has been accessed by connecting it to an
            # output of an upstream node with additional state variables.
            outpt._type_checked = False
        self._lzout = outputs
        outputs._node = self
        return outputs

    @property
    def splitter(self):
        if not self._state:
            return ()
        return self._state.splitter

    @property
    def combiner(self):
        if not self._state:
            return ()
        return self._state.combiner

    def _check_if_outputs_have_been_used(self, msg):
        used = []
        if self._lzout:
            for outpt_name, outpt_val in asdict(self._lzout).items():
                if outpt_val._type_checked:
                    used.append(outpt_name)
        if used:
            raise RuntimeError(
                f"Outputs {used} of {self} have already been accessed and therefore "
                + msg
            )

    def _set_state(self) -> None:
        # Add node name to state's splitter, combiner and container_ndim loaded from the def
        splitter = deepcopy(self._task._splitter)  # these can be modified in state
        combiner = deepcopy(self._task._combiner)  # these can be modified in state
        container_ndim = {}
        if splitter:
            splitter = add_name_splitter(splitter, self.name)
        if combiner:
            combiner = add_name_combiner(combiner, self.name)
        if self._task._container_ndim:
            for key, val in self._task._container_ndim.items():
                container_ndim[f"{self.name}.{key}"] = val
        other_states = self._get_upstream_states()
        if splitter or combiner or other_states:
            self._state = State(
                self.name,
                splitter=splitter,
                other_states=other_states,
                combiner=combiner,
                container_ndim=container_ndim,
            )
            if combiner:
                if not_split := [
                    c
                    for c in combiner
                    if not any(c in s for s in self.state.splitter_rpn) and "." not in c
                ]:
                    raise ValueError(
                        f"Combiner fields {not_split} for Node {self.name!r} are not in the "
                        f"splitter  {self.state.splitter_rpn}"
                    )
        else:
            self._state = None

    def _get_upstream_states(self) -> dict[str, tuple["State", list[str]]]:
        """Get the states of the upstream nodes that are connected to this node"""
        upstream_states = {}
        for inpt_name, val in self.input_values:
            if (
                isinstance(val, lazy.LazyOutField)
                and val._node.state
                and val._node.state.depth()
            ):
                node: Node = val._node
                # variables that are part of inner splitters should be treated as a containers
                if node.state and f"{node.name}.{val._field}" in node.state.splitter:
                    node.state._inner_container_ndim[f"{node.name}.{val._field}"] = 1
                # adding task_name: (task.state, [a field from the connection]
                if node.name not in upstream_states:
                    upstream_states[node.name] = (node.state, [val._field])
                else:
                    # if the task already exist in other_state,
                    # additional field name should be added to the list of fields
                    upstream_states[node.name][1].append(val._field)
        return upstream_states

        # else:
        #     # todo it never gets here
        #     breakpoint()
        #     inputs_dict = {inp: getattr(self.inputs, inp) for inp in self.input_names}
        #     return None, inputs_dict



import typing as ty
import inspect
from typing import dataclass_transform
import attrs
from pydra.compose import base
from pydra.compose.base import (
    ensure_field_objects,
    build_task_class,
    parse_doc_string,
    extract_function_inputs_and_outputs,
    check_explicit_fields_are_none,
    extract_fields_from_class,
)
from pydra.utils.general import attrs_values
from pydra.utils.typing import StateArray, is_lazy

if ty.TYPE_CHECKING:
    from pydra.engine.workflow import Workflow
    from pydra.engine.job import Job
    from pydra.engine.lazy import LazyOutField
    from pydra.engine.graph import DiGraph
    from pydra.engine.submitter import NodeExecution
    from pydra.environments.base import Environment
    from pydra.engine.hooks import TaskHooks


__all__ = ["define", "add", "this", "arg", "out", "Task", "Outputs", "cast"]


@attrs.define
class arg(base.Arg):
    """Argument of a workflow task

    Parameters
    ----------
    help: str
        A short description of the input field.
    default : Any, optional
        the default value for the argument
    allowed_values: list, optional
        List of allowed values for the field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    copy_mode: File.CopyMode, optional
        The mode of copying the file, by default it is File.CopyMode.any
    copy_collation: File.CopyCollation, optional
        The collation of the file, by default it is File.CopyCollation.any
    copy_ext_decomp: File.ExtensionDecomposition, optional
        The extension decomposition of the file, by default it is
        File.ExtensionDecomposition.single
    readonly: bool, optional
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}), by default
        it is False
    type: type, optional
        The type of the field, by default it is Any
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    lazy: bool, optional
        If True the input field is not required at construction time but is passed straight
        through to the tasks, by default it is False
    """

    pass


@attrs.define
class out(base.Out):
    """Output of a workflow task

    Parameters
    ----------
    name: str, optional
        The name of the field, used when specifying a list of fields instead of a mapping
        from name to field, by default it is None
    type: type, optional
        The type of the field, by default it is Any
    help: str, optional
        A short description of the input field.
    requires: list, optional
        Names of the inputs that are required together with the field.
    converter: callable, optional
        The converter for the field passed through to the attrs.field, by default it is None
    validator: callable | iterable[callable], optional
        The validator(s) for the field passed through to the attrs.field, by default it is None
    """

    pass


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(out,),
)
def outputs(wrapped):
    """Decorator to specify the output fields of a shell command is a dataclass-style type"""
    return wrapped


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(arg,),
)
def define(
    wrapped: type | ty.Callable | None = None,
    /,
    inputs: list[str | arg] | dict[str, arg | type] | None = None,
    outputs: list[str | out] | dict[str, out | type] | type | None = None,
    bases: ty.Sequence[type] = (),
    outputs_bases: ty.Sequence[type] = (),
    lazy: list[str] | None = None,
    auto_attribs: bool = True,
    name: str | None = None,
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]] = (),
) -> "Task":
    """
    Create an interface for a function or a class. Can be used either as a decorator on
    a constructor function or the "canonical" dataclass-form of a task.

    Parameters
    ----------
    wrapped : type | callable | None
        The function or class to create an interface for.
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The inputs to the function or class.
    outputs : list[str | Out] | dict[str, Out | type] | type | None
        The outputs of the function or class.
    auto_attribs : bool
        Whether to use auto_attribs mode when creating the class.
    name: str | None
        The name of the returned class
    xor: Sequence[str | None] | Sequence[Sequence[str | None]], optional
        Names of args that are exclusive mutually exclusive, which must include
        the name of the current field. If this list includes None, then none of the
        fields need to be set.

    Returns
    -------
    Task
        The interface for the function or class.
    """

    if lazy is None:
        lazy = []

    def make(wrapped: ty.Callable | type) -> Task:
        if inspect.isclass(wrapped):
            klass = wrapped
            constructor = klass.constructor
            class_name = klass.__name__
            check_explicit_fields_are_none(klass, inputs, outputs)
            parsed_inputs, parsed_outputs = extract_fields_from_class(
                Task,
                Outputs,
                klass,
                arg,
                out,
                auto_attribs,
                skip_fields=["constructor"],
            )
        else:
            if not inspect.isfunction(wrapped):
                raise ValueError(
                    f"wrapped must be a class or a function, not {wrapped!r}"
                )
            klass = None
            constructor = wrapped
            input_helps, output_helps = parse_doc_string(constructor.__doc__)
            inferred_inputs, inferred_outputs = extract_function_inputs_and_outputs(
                constructor, arg, inputs, outputs
            )

            class_name = constructor.__name__ if name is None else name

            parsed_inputs, parsed_outputs = ensure_field_objects(
                arg_type=arg,
                out_type=out,
                inputs=inferred_inputs,
                outputs=inferred_outputs,
                input_helps=input_helps,
                output_helps=output_helps,
            )

        if "constructor" in parsed_inputs:
            raise ValueError(
                "The argument 'constructor' is reserved and cannot be used as an "
                "argument name"
            )

        parsed_inputs["constructor"] = arg(
            name="constructor", type=ty.Callable, hash_eq=True, default=constructor
        )
        for inpt_name in lazy:
            parsed_inputs[inpt_name].lazy = True

        defn = build_task_class(
            Task,
            Outputs,
            parsed_inputs,
            parsed_outputs,
            name=class_name,
            klass=klass,
            bases=bases,
            outputs_bases=outputs_bases,
            xor=xor,
        )

        return defn

    if wrapped is not None:
        if not isinstance(wrapped, (ty.Callable, type)):
            raise ValueError(f"wrapped must be a class or a callable, not {wrapped!r}")
        return make(wrapped)
    return make


def this() -> "Workflow":
    """Get the workflow currently being constructed.

    Returns
    -------
    Workflow
        The workflow currently being constructed.
    """
    from pydra.engine.workflow import Workflow

    return Workflow.under_construction()


OutputsType = ty.TypeVar("OutputsType", bound="Outputs")


def add(
    task: "Task[OutputsType]",
    name: str | None = None,
    environment: "Environment | None" = None,
    hooks: "TaskHooks | None" = None,
) -> OutputsType:
    """Add a node to the workflow currently being constructed

    Parameters
    ----------
    task : Task
        The definition of the task to add to the workflow as a node
    name : str, optional
        The name of the node, by default it will be the name of the task
        class
    environment : Environment, optional
        The environment to run the task in, such as the Docker or Singularity container,
        by default it will be the "native"
    hooks : TaskHooks, optional
        The hooks to run before or after the task, by default no hooks will be run

    Returns
    -------
    Outputs
        The outputs of the node
    """
    return this().add(task, name=name, environment=environment, hooks=hooks)


U = ty.TypeVar("U")


def cast(field: ty.Any, new_type: type[U]) -> U:
    """Cast a lazy field to a new type. Note that the typing in the signature is a white
    lie, as the return field is actually a LazyField as placeholder for the object of
    type U.

    Parameters
    ----------
    field : LazyField[T]
        The field to cast
    new_type : type[U]
        The new type to cast the field to

    Returns
    -------
    LazyField[U]
        A copy of the lazy field with the new type
    """
    return attrs.evolve(
        field,
        type=new_type,
        cast_from=field._cast_from if field._cast_from else field._type,
    )


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class WorkflowOutputs(base.Outputs):

    @classmethod
    def _from_job(cls, job: "Job[WorkflowTask]") -> ty.Self:
        """Collect the outputs of a workflow job from the outputs of the nodes in the

        Parameters
        ----------
        job : Job[WorfklowDef]
            The job whose outputs are being collected.

        Returns
        -------
        outputs : Outputs
            The outputs of the job
        """

        workflow: "Workflow" = job.return_values["workflow"]
        exec_graph: "DiGraph[NodeExecution]" = job.return_values["exec_graph"]

        # Check for errors in any of the workflow nodes
        if errored := [n for n in exec_graph.nodes if n.errored]:
            errors = []
            for node in errored:
                for node_task in node.errored.values():
                    result = node_task.result()
                    if result.errors:
                        time_of_crash = result.errors["time of crash"]
                        error_message = "\n".join(result.errors["error message"])
                    else:
                        time_of_crash = "UNKNOWN-TIME"
                        error_message = "NOT RETRIEVED"
                    errors.append(
                        f"Job {node.name!r} failed @ {time_of_crash} running "
                        f"{node._task} with the following errors:\n{error_message}"
                        "\nTo inspect, please load the pickled job object from here: "
                        f"{result.cache_dir}/_job.pklz"
                    )
            raise RuntimeError(
                f"Workflow {job!r} failed with errors:\n\n" + "\n\n".join(errors)
            )

        # Retrieve values from the output fields
        values = {}
        lazy_field: LazyOutField
        for name, lazy_field in attrs_values(workflow.outputs).items():
            if is_lazy(lazy_field):
                val_out = lazy_field._get_value(workflow=workflow, graph=exec_graph)
            else:
                val_out = lazy_field  # handle non-lazy inputs that are passed through
            if isinstance(val_out, StateArray):
                val_out = list(val_out)  # implicitly combine state arrays
            values[name] = val_out

        # Set the values in the outputs object
        outputs = super()._from_job(job)
        outputs = attrs.evolve(outputs, **values)
        outputs._cache_dir = job.cache_dir
        return outputs


WorkflowOutputsType = ty.TypeVar("OutputType", bound=WorkflowOutputs)


@attrs.define(kw_only=True, auto_attribs=False, eq=False, repr=False)
class WorkflowTask(base.Task[WorkflowOutputsType]):

    _executor_name = "constructor"

    RESERVED_FIELD_NAMES = base.Task.RESERVED_FIELD_NAMES + ("construct",)

    _constructed = attrs.field(default=None, init=False, repr=False, eq=False)

    def _run(self, job: "Job[WorkflowTask]", rerun: bool) -> None:
        """Run the workflow."""
        job.submitter.expand_workflow(job, rerun)

    async def _run_async(self, job: "Job[WorkflowTask]", rerun: bool) -> None:
        """Run the workflow asynchronously."""
        await job.submitter.expand_workflow_async(job, rerun)

    def construct(self) -> "Workflow":
        from pydra.engine.workflow import Workflow

        if self._constructed is not None:
            return self._constructed
        self._constructed = Workflow.construct(self)
        return self._constructed


# Alias WorkflowTask to Task so we can refer to it as workflow.Task
Task = WorkflowTask
Outputs = WorkflowOutputs



"""Basic processing graph elements."""

import json
import logging
import os
import inspect
import sys
import asyncio
from pathlib import Path
import typing as ty
from uuid import uuid4
import shutil
from traceback import format_exception
import attr
import cloudpickle as cp
from pydra.compose.base import Task
from pydra.utils.hash import hash_function
from filelock import SoftFileLock, Timeout
from datetime import datetime
from fileformats.core import FileSet
from pydra.engine.hooks import TaskHooks
from pydra.engine.result import (
    RuntimeSpec,
    Result,
    record_error,
)
from pydra.utils.general import (
    attrs_values,
    attrs_fields,
    get_fields,
    ensure_list,
    is_workflow,
)
from pydra.utils.typing import is_lazy
from pydra.engine.result import load_result, save
from pydra.utils.typing import copy_nested_files
from pydra.compose.shell.templating import template_update
from pydra.utils.messenger import AuditFlag
from pydra.environments.base import Environment

logger = logging.getLogger("pydra")

develop = False

if ty.TYPE_CHECKING:
    from pydra.engine.submitter import Submitter
    from pydra.compose.base import Arg

TaskType = ty.TypeVar("TaskType", bound=Task)


class Job(ty.Generic[TaskType]):
    """
    A base structure for the nodes in the processing graph.

    Tasks are a generic compute step from which both elementary tasks and
    :class:`Workflow` instances inherit.

    """

    _api_version: str = "0.0.1"  # Should generally not be touched by subclasses
    _etelemetry_version_data = None  # class variable to store etelemetry information
    _version: str  # Version of tool being wrapped
    _task_version: ty.Optional[str] = None
    # Job writers encouraged to define and increment when implementation changes sufficiently
    _input_sets = None  # Dictionaries of predefined input settings

    audit_flags: AuditFlag = AuditFlag.NONE
    """What to audit -- available flags: :class:`~pydra.utils.messenger.AuditFlag`."""

    _can_resume = False  # Does the job allow resuming from previous state
    _redirect_x = False  # Whether an X session should be created/directed

    _runtime_requirements = RuntimeSpec()
    _runtime_hints = None

    _cache_root = None  # Working directory in which to operate
    _references = None  # List of references for a job

    name: str
    task: TaskType
    submitter: "Submitter | None"
    environment: "Environment | None"
    state_index: int
    bindings: dict[str, ty.Any] | None = None  # Bindings for the job environment

    _inputs: dict[str, ty.Any] | None = None
    _run_start_time: datetime | None

    def __init__(
        self,
        task: TaskType,
        submitter: "Submitter",
        name: str,
        environment: "Environment | None" = None,
        state_index: int | None = None,
        hooks: TaskHooks | None = None,
    ):
        """
        Initialize a job.

        Jobs allow for caching (retrieving a previous result of the same
        task and inputs), and concurrent execution.
        Running tasks follows a decision flow:

            1. Check whether prior cache exists --
               if ``True``, return cached result
            2. Check whether other process is running this job --
               wait if ``True``:
               a. Finishes (with or without exception) -> return result
               b. Gets killed -> restart
            3. No cache or other process -> start
            4. Two or more concurrent new processes get to start
        """

        if not isinstance(task, Task):
            raise ValueError(f"Job task ({task!r}) must be a Task, not {type(task)}")
        # Check that the task is fully resolved and ready to run
        task._check_resolved()
        task._check_rules()
        self.task = task
        # We save the submitter is the task is a workflow otherwise we don't
        # so the job can be pickled
        self.submitter = submitter
        self.environment = (
            environment if environment is not None else submitter.environment
        )
        self.name = name
        self.state_index = state_index

        self.return_values = {}
        self._result = {}
        # flag that says if node finished all jobs
        self._done = False
        if self._input_sets is None:
            self._input_sets = {}

        self.allow_cache_override = True
        self._checksum = None
        self._uid = uuid4().hex
        self.hooks = hooks if hooks is not None else TaskHooks()
        self._errored = False
        self._lzout = None

        # Save the submitter attributes needed to run the job later
        self.audit = submitter.audit
        self.cache_root = submitter.cache_root
        self.all_caches = submitter.readonly_caches
        self._run_start_time = None

    @property
    def cache_root(self):
        return self._cache_root

    @property
    def is_async(self) -> bool:
        """Check to see if the job should be run asynchronously."""
        return self.submitter.worker.is_async and is_workflow(self.task)

    @cache_root.setter
    def cache_root(self, path: os.PathLike):
        self._cache_root = Path(path)

    @property
    def all_caches(self):
        """Get the list of cache sources."""
        return ensure_list(self.cache_root) + self._readonly_caches

    @all_caches.setter
    def all_caches(self, locations):
        if locations is not None:
            self._readonly_caches = [Path(loc) for loc in ensure_list(locations)]
        else:
            self._readonly_caches = []

    def __str__(self):
        return self.name

    def __getstate__(self):
        state = self.__dict__.copy()
        state["task"] = cp.dumps(state["task"])
        return state

    def __setstate__(self, state):
        state["task"] = cp.loads(state["task"])
        self.__dict__.update(state)

    @property
    def errored(self):
        """Check if the job has raised an error"""
        return self._errored

    @property
    def checksum(self):
        """Calculates the unique checksum of the job.
        Used to create specific directory name for job that are run;
        and to create nodes checksums needed for graph checksums
        (before the tasks have inputs etc.)
        """
        if self._checksum is not None:
            return self._checksum
        self._checksum = self.task._checksum
        return self._checksum

    @property
    def lockfile(self):
        return self.cache_dir.with_suffix(".lock")

    @property
    def uid(self):
        """the unique id number for the job
        It will be used to create unique names for slurm scripts etc.
        without a need to run checksum
        """
        return self._uid

    @property
    def output_names(self):
        """Get the names of the outputs from the job's output_spec"""
        return [f.name for f in attr.fields(self.task.Outputs)]

    @property
    def can_resume(self):
        """Whether the job accepts checkpoint-restart."""
        return self._can_resume

    @property
    def cache_dir(self):
        """Get the filesystem path where outputs will be written."""
        return self.cache_root / self.checksum

    @property
    def inputs(self) -> dict[str, ty.Any]:
        """Resolve any template inputs of the job ahead of its execution:

        - links/copies upstream files and directories into the destination tasks
          working directory as required select state array values corresponding to
          state index (it will try to leave them where they are unless specified or
          they are on different file systems)
        - resolve template values (e.g. output_file_template)
        - deepcopy all inputs to guard against in-place changes during the job's
          execution (they will be replaced after the job's execution with the
          original inputs to ensure the tasks checksums are consistent)
        """
        if self._inputs is not None:
            return self._inputs

        from pydra.utils.typing import TypeParser

        self._inputs = {
            k: v for k, v in attrs_values(self.task).items() if not k.startswith("_")
        }
        map_copyfiles = {}
        fld: "Arg"
        for fld in get_fields(self.task):
            name = fld.name
            value = self._inputs[name]
            if value and TypeParser.contains_type(FileSet, fld.type):
                copied_value = copy_nested_files(
                    value=value,
                    dest_dir=self.cache_dir,
                    mode=fld.copy_mode,
                    collation=fld.copy_collation,
                    supported_modes=self.SUPPORTED_COPY_MODES,
                )
                if value is not copied_value:
                    map_copyfiles[name] = copied_value
        self._inputs.update(
            template_update(
                self.task, cache_dir=self.cache_dir, map_copyfiles=map_copyfiles
            )
        )
        return self._inputs

    def _populate_filesystem(self):
        """
        Invoked immediately after the lockfile is generated, this function:
        - Creates the cache file
        - Clears existing outputs if `can_resume` is False
        - Generates a fresh output directory

        Created as an attempt to simplify overlapping `Job`|`Workflow` behaviors.
        """
        # adding info file with the checksum in case the job was cancelled
        # and the lockfile has to be removed
        with open(self.cache_root / f"{self.uid}_info.json", "w") as jsonfile:
            json.dump({"checksum": self.checksum}, jsonfile)
        if not self.can_resume and self.cache_dir.exists():
            shutil.rmtree(self.cache_dir)
        self.cache_dir.mkdir(parents=False, exist_ok=self.can_resume)
        # Save job pkl into the output directory for future reference
        save(self.cache_dir, job=self)

    def run(self, rerun: bool = False):
        """Prepare the job working directory, execute the task, and save the
        results.

        Parameters
        ----------
        rerun : bool
            If True, the job will be re-run even if a result already exists. Will
            propagated to all tasks within workflow tasks.
        """
        # TODO: After these changes have been merged, will refactor this function and
        # run_async to use common helper methods for pre/post run tasks

        # checking if the task is fully resolved and ready to run
        self.hooks.pre_run(self)
        logger.debug(
            "'%s' is attempting to acquire lock on %s", self.name, self.lockfile
        )
        with SoftFileLock(self.lockfile):
            if not (rerun):
                result = self.result()
                if result is not None and not result.errored:
                    return result
            cwd = os.getcwd()
            self._populate_filesystem()
            os.chdir(self.cache_dir)
            result = Result(
                outputs=None,
                runtime=None,
                errored=False,
                cache_dir=self.cache_dir,
                task=self.task,
            )
            self.hooks.pre_run_task(self)
            self.audit.start_audit(odir=self.cache_dir)
            if self.audit.audit_check(AuditFlag.PROV):
                self.audit.audit_task(job=self)
            try:
                self.audit.monitor()
                self.task._run(self, rerun)
                result.outputs = self.task.Outputs._from_job(self)
            except Exception:
                etype, eval, etr = sys.exc_info()
                traceback = format_exception(etype, eval, etr)
                record_error(self.cache_dir, error=traceback)
                result.errored = True
                raise
            finally:
                self.hooks.post_run_task(self, result)
                self.audit.finalize_audit(result=result)
                save(self.cache_dir, result=result, job=self)
                # removing the additional file with the checksum
                (self.cache_root / f"{self.uid}_info.json").unlink()
                os.chdir(cwd)
        self.hooks.post_run(self, result)
        # Check for any changes to the input hashes that have occurred during the execution
        # of the job
        self._check_for_hash_changes()
        return result

    async def run_async(self, rerun: bool = False) -> Result:
        """Prepare the job working directory, execute the task asynchronously,
        and save the results. NB: only workflows are run asynchronously at the moment.

        Parameters
        ----------
        rerun : bool
            If True, the job will be re-run even if a result already exists. Will
            propagated to all tasks within workflow tasks.
        """
        # checking if the task is fully resolved and ready to run
        self.hooks.pre_run(self)
        logger.debug(
            "'%s' is attempting to acquire lock on %s", self.name, self.lockfile
        )
        async with PydraFileLock(self.lockfile):
            if not rerun:
                result = self.result()
                if result is not None and not result.errored:
                    return result
            cwd = os.getcwd()
            self._populate_filesystem()
            result = Result(
                outputs=None,
                runtime=None,
                errored=False,
                cache_dir=self.cache_dir,
                task=self.task,
            )
            self.hooks.pre_run_task(self)
            self.audit.start_audit(odir=self.cache_dir)
            try:
                self.audit.monitor()
                await self.task._run_async(self, rerun)
                result.outputs = self.task.Outputs._from_job(self)
            except Exception:
                etype, eval, etr = sys.exc_info()
                traceback = format_exception(etype, eval, etr)
                record_error(self.cache_dir, error=traceback)
                result.errored = True
                self._errored = True
                raise
            finally:
                self.hooks.post_run_task(self, result)
                self.audit.finalize_audit(result=result)
                save(self.cache_dir, result=result, job=self)
                # removing the additional file with the checksum
                (self.cache_root / f"{self.uid}_info.json").unlink()
                os.chdir(cwd)
        self.hooks.post_run(self, result)
        # Check for any changes to the input hashes that have occurred during the execution
        # of the job
        self._check_for_hash_changes()
        return result

    def pickle_task(self):
        """Pickling the tasks with full inputs"""
        pkl_files = self.cache_root / "pkl_files"
        pkl_files.mkdir(exist_ok=True, parents=True)
        task_main_path = pkl_files / f"{self.name}_{self.uid}_job.pklz"
        save(task_path=pkl_files, job=self, name_prefix=f"{self.name}_{self.uid}")
        return task_main_path

    @property
    def done(self):
        """Check whether the tasks has been finalized and all outputs are stored."""
        # if any of the field is lazy, there is no need to check results
        if has_lazy(self.task):
            return False
        _result = self.result()
        if _result:
            if _result.errored:
                self._errored = True
                raise ValueError(f"Job {self.name!r} failed")
            else:
                return True
        return False

    @property
    def run_start_time(self) -> datetime | None:
        """Check whether the job is currently running."""
        if self._run_start_time is not None:
            return self._run_start_time
        try:
            stat = self.lockfile.stat()
        except FileNotFoundError:
            return None
        self._run_start_time = datetime.fromtimestamp(stat.st_ctime)
        return self._run_start_time

    def _combined_output(self, return_inputs=False):
        combined_results = []
        for gr, ind_l in self.state.final_combined_ind_mapping.items():
            combined_results_gr = []
            for ind in ind_l:
                result = load_result(self.checksum_states(ind), self.all_caches)
                if result is None:
                    return None
                if return_inputs is True or return_inputs == "val":
                    result = (self.state.states_val[ind], result)
                elif return_inputs is True or return_inputs == "ind":
                    result = (self.state.states_ind[ind], result)
                combined_results_gr.append(result)
            combined_results.append(combined_results_gr)
        if len(combined_results) == 1 and self.state.splitter_rpn_final == []:
            # in case it's full combiner, removing the nested structure
            return combined_results[0]
        else:
            return combined_results

    def result(self, return_inputs=False):
        """
        Retrieve the outcomes of this particular job.

        Parameters
        ----------
        state_index : :obj: `int`
            index of the element for job with splitter and multiple states
        return_inputs : :obj: `bool`, :obj:`str`
            if True or "val" result is returned together with values of the input fields,
            if "ind" result is returned together with indices of the input fields

        Returns
        -------
        result : Result
            the result of the job
        """
        if self.errored:
            return Result(
                outputs=None,
                runtime=None,
                errored=True,
                cache_dir=self.cache_dir,
                task=self.task,
            )

        checksum = self.checksum
        result = load_result(checksum, self.all_caches)
        if result and result.errored:
            self._errored = True
        if return_inputs is True or return_inputs == "val":
            inputs_val = {
                f"{self.name}.{inp}": getattr(self.task, inp)
                for inp in self.input_names
            }
            return (inputs_val, result)
        elif return_inputs == "ind":
            inputs_ind = {f"{self.name}.{inp}": None for inp in self.input_names}
            return (inputs_ind, result)
        else:
            return result

    def _check_for_hash_changes(self):
        hash_changes = self.task._hash_changes()
        details = ""
        for changed in hash_changes:
            field = getattr(attr.fields(type(self.task)), changed)
            hash_function(getattr(self.task, changed))
            val = getattr(self.task, changed)
            field_type = type(val)
            if inspect.isclass(field.type) and issubclass(field.type, FileSet):
                details += (
                    f"- {changed}: value passed to the {field.type} field is of type "
                    f"{field_type} ('{val}'). If it is intended to contain output data "
                    "then the type of the field in the interface class should be changed "
                    "to `pathlib.Path`. Otherwise, if the field is intended to be an "
                    "input field but it gets altered by the job in some way, then the "
                    "'copyfile' flag should be set to 'copy' in the field metadata of "
                    "the job interface class so copies of the files/directories in it "
                    "are passed to the job instead.\n"
                )
            else:
                details += (
                    f"- {changed}: the {field_type} object passed to the {field.type}"
                    f"field appears to have an unstable hash. This could be due to "
                    "a stochastic/non-thread-safe attribute(s) of the object\n\n"
                    f'A "bytes_repr" method for {field.type!r} can be implemented to '
                    "bespoke hashing methods based only on the stable attributes for "
                    f"the `{field_type.__module__}.{field_type.__name__}` type. "
                    f"See pydra/utils/hash.py for examples. Value: {val}\n"
                )
        if hash_changes:
            raise RuntimeError(
                f"Input field hashes have changed during the execution of the "
                f"'{self.name}' job of {type(self)} type.\n\n{details}"
            )
        logger.debug(
            "Input values and hashes for '%s' %s node:\n%s\n%s",
            self.name,
            type(self).__name__,
            self.task,
            self.task._hashes,
        )

    def _write_notebook(self):
        """Writes a notebook into the"""
        raise NotImplementedError

    SUPPORTED_COPY_MODES = FileSet.CopyMode.any
    DEFAULT_COPY_COLLATION = FileSet.CopyCollation.any


def has_lazy(obj):
    """Check whether an object has lazy fields."""
    for f in attrs_fields(obj):
        if is_lazy(getattr(obj, f.name)):
            return True
    return False


class PydraFileLock:
    """Wrapper for filelock's SoftFileLock that makes it work with asyncio."""

    def __init__(self, lockfile):
        self.lockfile = lockfile
        self.timeout = 0.1

    async def __aenter__(self):
        lock = SoftFileLock(self.lockfile)
        acquired_lock = False
        while not acquired_lock:
            try:
                lock.acquire(timeout=0)
                acquired_lock = True
            except Timeout:
                await asyncio.sleep(self.timeout)
                if self.timeout <= 2:
                    self.timeout = self.timeout * 2
        self.lock = lock
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.lock.release()
        return None


def load_and_run(job_pkl: Path, rerun: bool = False) -> Path:
    """
    loading a job from a pickle file, settings proper input
    and running the job

    Parameters
    ----------
    job_pkl : :obj:`Path`
        The path to pickled job file

    Returns
    -------
    resultfile : :obj:`Path`
        The path to the pickled result file
    """

    try:
        job: Job[TaskType] = load_job(job_pkl=job_pkl)
    except Exception:
        if job_pkl.parent.exists():
            etype, eval, etr = sys.exc_info()
            traceback = format_exception(etype, eval, etr)
            errorfile = record_error(job_pkl.parent, error=traceback)
            result = Result(output=None, runtime=None, errored=True, task=None)
            save(job_pkl.parent, result=result)
        raise

    resultfile = job.cache_dir / "_result.pklz"
    try:
        if job.is_async:
            job.submitter.submit(job, rerun=rerun)
        else:
            job.run(rerun=rerun)
    except Exception as e:
        # creating result and error files if missing
        errorfile = job.cache_dir / "_error.pklz"
        if not errorfile.exists():  # not sure if this is needed
            etype, eval, etr = sys.exc_info()
            traceback = format_exception(etype, eval, etr)
            errorfile = record_error(job.cache_dir, error=traceback)
        if not resultfile.exists():  # not sure if this is needed
            result = Result(output=None, runtime=None, errored=True, task=None)
            save(job.cache_dir, result=result)
        e.add_note(f" full crash report is here: {errorfile}")
        raise
    return resultfile


def load_job(job_pkl: os.PathLike) -> "Job[TaskType]":
    """loading a job from a pickle file, settings proper input for the specific ind"""
    with open(job_pkl, "rb") as fp:
        job = cp.load(fp)
    return job



import typing as ty
import re
import os
import inspect
from copy import copy
from pathlib import Path
from fileformats.generic import FileSet
from pydra.utils.general import attrs_values, get_fields
from pydra.utils.typing import is_lazy
from . import field

if ty.TYPE_CHECKING:
    from . import Task


def template_update(
    task,
    cache_dir: Path | None = None,
    map_copyfiles: dict[str, Path] | None = None,
):
    """
    Update all templates that are present in the input task.

    Should be run when all inputs used in the templates are already set.

    """

    values = attrs_values(task)
    if map_copyfiles is not None:
        values.update(map_copyfiles)

    # Collect templated inputs for which all requirements are satisfied.
    fields_templ = [
        fld
        for fld in get_fields(task)
        if isinstance(fld, field.outarg)
        and fld.path_template
        and getattr(task, fld.name)
        and all(req.satisfied(task) for req in fld.requires)
    ]

    dict_mod = {}
    for fld in fields_templ:
        dict_mod[fld.name] = template_update_single(
            fld=fld,
            task=task,
            values=values,
            cache_dir=cache_dir,
        )
    # adding elements from map_copyfiles to fields with templates
    if map_copyfiles:
        dict_mod.update(map_copyfiles)
    return dict_mod


def template_update_single(
    fld: "field.outarg",
    task: "Task",
    values: dict[str, ty.Any] = None,
    cache_dir: Path | None = None,
    spec_type: str = "input",
) -> Path | list[Path | None] | None:
    """Update a single template from the input_spec or output_spec
    based on the value from inputs_dict
    (checking the types of the fields, that have "output_file_template)"
    """
    # if input_dict_st with state specific value is not available,
    # the dictionary will be created from inputs object
    from pydra.utils.typing import TypeParser, OUTPUT_TEMPLATE_TYPES  # noqa

    if values is None:
        values = attrs_values(task)

    if spec_type == "input":
        field_value = values[fld.name]
        if isinstance(field_value, bool) and fld.type in (Path, str):
            raise TypeError(
                f"type of '{fld.name}' is Path, consider using Union[Path, bool]"
            )
        if field_value is not None and not is_lazy(field_value):
            field_value = TypeParser(ty.Union[OUTPUT_TEMPLATE_TYPES])(field_value)
    elif spec_type == "output":
        if not TypeParser.contains_type(FileSet, fld.type):
            raise TypeError(
                f"output {fld.name} should be file-system object, but {fld.type} "
                "set as the type"
            )
    else:
        raise TypeError(f"spec_type can be input or output, but {spec_type} provided")
    # for inputs that the value is set (so the template is ignored)
    if spec_type == "input":
        if isinstance(field_value, (Path, list)):
            return field_value
        if field_value is False:
            # if input fld is set to False, the fld shouldn't be used (setting NOTHING)
            return None
    # inputs_dict[fld.name] is True or spec_type is output
    value = _template_formatting(fld, task, values)
    if cache_dir and value is not None:
        # changing path so it is in the cache_dir
        # should be converted to str, it is also used for input fields that should be str
        if type(value) is list:
            value = [cache_dir / val.name for val in value]
        else:
            value = cache_dir / value.name
    return value


def _template_formatting(
    fld: "field.arg", task: "Task", values: dict[str, ty.Any]
) -> Path | list[Path] | None:
    """Formatting the fld template based on the values from inputs.
    Taking into account that the fld with a template can be a MultiOutputFile
    and the fld values needed in the template can be a list -
    returning a list of formatted templates in that case.
    Allowing for multiple input values used in the template as longs as
    there is no more than one file (i.e. File, PathLike or string with extensions)

    Parameters
    ----------
    fld : pydra.utils.general.Field
        field with a template
    task : pydra.compose.shell.Task
        the task
    values : dict
        dictionary with values from inputs object

    Returns
    -------
    formatted : Path or list[Path | None] or None
        formatted template
    """
    # if a template is a function it has to be run first with the inputs as the only arg
    template = fld.path_template
    if callable(template):
        template = template(task)

    # as default, we assume that keep_extension is True
    if isinstance(template, (tuple, list)):
        formatted = [_single_template_formatting(fld, t, values) for t in template]
        if any([val is None for val in formatted]):
            return None
    else:
        assert isinstance(template, str)
        formatted = _single_template_formatting(fld, template, values)
    return formatted


def _single_template_formatting(
    fld: "field.outarg",
    template: str,
    values: dict[str, ty.Any],
) -> Path | None:
    from pydra.utils.typing import MultiInputObj, MultiOutputFile

    inp_fields = re.findall(r"{\w+}", template)
    inp_fields_fl = re.findall(r"{\w+:[0-9.]+f}", template)
    inp_fields += [re.sub(":[0-9.]+f", "", el) for el in inp_fields_fl]

    # FIXME: This would be a better solution, and would allow you to explicitly specify
    # whether you want to use the extension of the input file or not, by referencing
    # the "ext" attribute of the input file. However, this would require a change in the
    # way the element formatting is done
    #
    # inp_fields = set(re.findall(r"{(\w+)(?:\.\w+)?(?::[0-9.]+f)?}", template))

    if len(inp_fields) == 0:
        return Path(template)

    val_dict = {}
    file_template = None

    for inp_fld in inp_fields:
        fld_name = inp_fld[1:-1]  # extracting the name form {field_name}
        if fld_name not in values:
            raise AttributeError(f"{fld_name} is not provided in the input")
        fld_value = values[fld_name]
        if fld_value is None:
            # if value is NOTHING, nothing should be added to the command
            return None
        # checking for fields that can be treated as a file:
        # have type File, or value that is path like (including str with extensions)
        if isinstance(fld_value, os.PathLike):
            if file_template:
                raise Exception(
                    f"can't have multiple paths in {fld.name} template,"
                    f" but {template} provided"
                )
            else:
                file_template = (fld_name, fld_value)
        else:
            val_dict[fld_name] = fld_value

    # if field is MultiOutputFile and some elements from val_dict are lists,
    # each element of the list should be used separately in the template
    # and return a list with formatted values
    if fld.type is MultiOutputFile and any(
        [isinstance(el, (list, MultiInputObj)) for el in val_dict.values()]
    ):
        # all fields that are lists
        keys_list = [
            k for k, el in val_dict.items() if isinstance(el, (list, MultiInputObj))
        ]
        if any(
            [len(val_dict[key]) != len(val_dict[keys_list[0]]) for key in keys_list[1:]]
        ):
            raise Exception(
                f"all fields used in {fld.name} template have to have the same length"
                f" or be a single value"
            )
        formatted_value = []
        for ii in range(len(val_dict[keys_list[0]])):
            val_dict_el = copy(val_dict)
            # updating values to a single element from the list
            for key in keys_list:
                val_dict_el[key] = val_dict[key][ii]

            formatted_value.append(
                _element_formatting(
                    template,
                    val_dict_el,
                    file_template,
                    keep_extension=fld.keep_extension,
                )
            )
    else:
        formatted_value = _element_formatting(
            template, val_dict, file_template, keep_extension=fld.keep_extension
        )
    if isinstance(formatted_value, list):
        return [Path(val) for val in formatted_value]
    elif isinstance(formatted_value, str):
        return Path(formatted_value)
    return None


def _element_formatting(
    template: str,
    values_template_dict: dict[str, ty.Any],
    file_template: str,
    keep_extension: bool,
):
    """Formatting a single template for a single element (if a list).
    Taking into account that a file used in the template (file_template)
    and the template itself could have file extensions
    (assuming that if template has extension, the field value extension is removed,
    if field has extension, and no template extension, than it is moved to the end).
    For values_template_dict the simple formatting can be used (no file values inside)
    """
    if file_template:
        fld_name_file, fld_value_file = file_template
        # splitting the filename for name and extension,
        # the final value used for formatting depends on the template and keep_extension flag
        name, *ext = Path(fld_value_file).name.split(".", maxsplit=1)
        filename = str(Path(fld_value_file).parent / name)
        # updating values_template_dic with the name of file
        values_template_dict[fld_name_file] = filename
        # if keep_extension is False, the extensions are removed
        if keep_extension is False:
            ext = []
    else:
        ext = []

    # if file_template is at the end of the template, the simplest formatting should work
    if file_template and template.endswith(f"{{{fld_name_file}}}"):
        # recreating fld_value with the updated extension
        values_template_dict[fld_name_file] = ".".join([filename] + ext)
        formatted_value = template.format(**values_template_dict)
    # file_template provided, but the template doesn't have its own extension
    elif file_template and "." not in template:
        # if the fld_value_file has extension, it will be moved to the end
        formatted_value = ".".join([template.format(**values_template_dict)] + ext)
    # template has its own extension or no file_template provided
    # the simplest formatting, if file_template is provided it's used without the extension
    else:
        formatted_value = template.format(**values_template_dict)
    return formatted_value


def parse_format_string(fmtstr: str) -> set[str]:
    """Parse a argstr format string and return all keywords used in it."""
    identifier = r"[a-zA-Z_]\w*"
    attribute = rf"\.{identifier}"
    item = r"\[\w+\]"
    # Example: var.attrs[key][0].attr2 (capture "var")
    field_with_lookups = (
        f"({identifier})(?:{attribute}|{item})*"  # Capture only the keyword
    )
    conversion = "(?:!r|!s)"
    nobrace = "[^{}]*"
    # Example: 0{pads[hex]}x (capture "pads")
    fmtspec = f"{nobrace}(?:{{({identifier}){nobrace}}}{nobrace})?"  # Capture keywords in definition
    full_field = f"{{{field_with_lookups}{conversion}?(?::{fmtspec})?}}"

    all_keywords = re.findall(full_field, fmtstr)
    return set().union(*all_keywords) - {""}


def fields_in_formatter(formatter: str | ty.Callable[..., str]) -> set[str]:
    """Extract all field names from a formatter string or function."""
    if isinstance(formatter, str):
        return parse_format_string(formatter)
    elif isinstance(formatter, ty.Sequence):
        return set().union(*[fields_in_formatter(f) for f in formatter])
    elif isinstance(formatter, ty.Callable):
        return set(inspect.signature(formatter).parameters.keys())
    else:
        raise ValueError(f"Unsupported formatter type: {type(formatter)} ({formatter})")


def argstr_formatting(argstr: str, values: dict[str, ty.Any]):
    """formatting argstr that have form {field_name},
    using values from inputs and updating with value_update if provided
    """
    # if there is a value that has to be updated (e.g. single value from a list)
    # getting all fields that should be formatted, i.e. {field_name}, ...
    inp_fields = parse_format_string(argstr)
    # formatting string based on the val_dict
    argstr_formatted = argstr.format(**{n: values.get(n, "") for n in inp_fields})
    # removing extra commas and spaces after removing the field that have NOTHING
    argstr_formatted = (
        argstr_formatted.replace("[ ", "[")
        .replace(" ]", "]")
        .replace("[,", "[")
        .replace(",]", "]")
        .strip()
    )
    return argstr_formatted



"""Administrative support for the engine framework."""

from pathlib import Path
import inspect
import sys
import typing as ty
from collections.abc import Mapping, Collection
from copy import copy
import re
import attrs
import ast
import tempfile
import importlib
import types
import sysconfig
import platformdirs
import builtins
import pkgutil
import logging
import pydra.scripts
from ._version import __version__


logger = logging.getLogger("pydra")
if ty.TYPE_CHECKING:
    from pydra.compose.base import Task, Field  # noqa
    from pydra.compose import workflow


SCRIPTS_DIR = Path(pydra.scripts.__file__).parent

PYDRA_ATTR_METADATA = "__PYDRA_METADATA__"

TaskType = ty.TypeVar("TaskType", bound="Task")


user_cache_root = Path(
    platformdirs.user_cache_dir(
        appname="pydra",
        appauthor="nipype",
        version=__version__,
    )
)

default_run_cache_root = user_cache_root / "run-cache"


def add_exc_note(e: Exception, note: str) -> Exception:
    """Adds a note to an exception in a Python <3.11 compatible way

    Parameters
    ----------
    e : Exception
        the exception to add the note to
    note : str
        the note to add

    Returns
    -------
    Exception
        returns the exception again
    """
    if hasattr(e, "add_note"):
        e.add_note(note)
    else:
        e.args = (e.args[0] + "\n" + note,)
    return e


def exc_info_matches(exc_info, match, regex=False):
    if exc_info.value.__cause__ is not None:
        msg = str(exc_info.value.__cause__)
    else:
        msg = str(exc_info.value)
    if regex:
        return re.match(".*" + match, msg)
    else:
        return match in msg


def get_undefined_symbols(
    func, exclude_signature_type_hints: bool = False, ignore_decorator: bool = False
):
    """
    Check the source code of a function and detect any symbols that aren't defined in its scope.

    Parameters
    ----------
    func : callable
        The function to analyze.

    Returns
    -------
    set
        A set of undefined symbols.
    """
    # Get the source code of the function
    source = inspect.getsource(func)

    # De-indent the source code if required
    indent = re.match(r"^\s*", source).group()
    source = ("\n" + source).replace("\n" + indent, "\n")

    if ignore_decorator:
        # Remove the decorator from the source code, i.e. everything before the first
        # unindented 'def ' keyword.
        source = re.match(
            r"(.*\n)(def .*)", "\n" + source, flags=re.MULTILINE | re.DOTALL
        ).group(2)

    # Parse the source code into an AST
    tree = ast.parse(source)

    # Define a visitor class to traverse the AST
    class SymbolVisitor(ast.NodeVisitor):

        def __init__(self):
            # Initialize sets to track defined and used symbols
            self.defined_symbols = set()
            self.used_symbols = set()

        def visit_FunctionDef(self, node):
            # Add function arguments to defined symbols
            for arg in node.args.args:
                self.defined_symbols.add(arg.arg)
            if exclude_signature_type_hints:
                # Exclude type hints from the defined symbols
                type_hints_visitor = SymbolVisitor()
                if node.returns:
                    type_hints_visitor.visit(node.returns)
                for arg in node.args.args:
                    if arg.annotation:
                        type_hints_visitor.visit(arg.annotation)
                type_hint_symbols = type_hints_visitor.used_symbols - self.used_symbols
            self.generic_visit(node)
            if exclude_signature_type_hints:
                # Remove type hints from the used symbols
                self.used_symbols -= type_hint_symbols

        def visit_Assign(self, node):
            # Add assigned variables to defined symbols
            for target in node.targets:
                if isinstance(target, ast.Name):
                    self.defined_symbols.add(target.id)
            self.generic_visit(node)

        def visit_Name(self, node):
            # Add all variable names to used symbols
            if isinstance(node.ctx, ast.Load):
                self.used_symbols.add(node.id)
            self.generic_visit(node)

        @property
        def undefined_symbols(self):
            return self.used_symbols - self.defined_symbols - get_builtin_type_names()

    # Create a visitor instance and visit the AST
    visitor = SymbolVisitor()
    visitor.visit(tree)

    return visitor.undefined_symbols


def get_builtin_type_names():
    """
    Get a list of built-in object type names in Python.

    Returns
    -------
    set
        A set of built-in object type names.
    """
    return set(name for name, obj in vars(builtins).items() if isinstance(obj, type))


def in_stdlib(obj: types.FunctionType | type) -> str | bool:
    """Check if a type is in the standard library and return the name of the module if
    so."""
    module = inspect.getmodule(obj)
    if module is None:
        return False
    if module.__name__.startswith("builtins"):
        return "builtins"
    if module.__name__ == "types" and obj.__name__ not in dir(types):
        return False
    toplevel = module.__name__.split(".")[0]
    if toplevel in STDLIB_MODULES:
        return toplevel
    return False


def _stdlib_modules() -> frozenset[str]:
    """List all standard library modules."""
    std_lib_modules = set(sys.builtin_module_names)
    std_lib_path = sysconfig.get_path("stdlib")
    std_lib_modules.update(m[1] for m in pkgutil.iter_modules([std_lib_path]))
    return frozenset(std_lib_modules)


STDLIB_MODULES: frozenset[str] = _stdlib_modules()

# Example usage:
# print(list_standard_library_modules())


def plot_workflow(
    workflow_task: "workflow.Task",
    out_dir: Path,
    plot_type: str = "simple",
    export: ty.Sequence[str] | None = None,
    name: str | None = None,
    lazy: ty.Sequence[str] | ty.Set[str] | None = None,
) -> Path | tuple[Path, list[Path]]:
    """creating a graph - dotfile and optionally exporting to other formats"""
    from pydra.engine.workflow import Workflow

    if inspect.isclass(workflow_task):
        workflow_task = workflow_task()

    # Create output directory
    out_dir.mkdir(parents=True, exist_ok=True)

    if lazy is None:
        lazy = [n for n, v in attrs_values(workflow_task).items() if v is attrs.NOTHING]

    # Construct the workflow object with all of the fields lazy
    wf = Workflow.construct(workflow_task, lazy=lazy)

    if not name:
        name = f"graph_{type(workflow_task).__name__}"
    if plot_type == "simple":
        graph = wf.graph()
        dotfile = graph.create_dotfile_simple(outdir=out_dir, name=name)
    elif plot_type == "nested":
        graph = wf.graph()
        dotfile = graph.create_dotfile_nested(outdir=out_dir, name=name)
    elif plot_type == "detailed":
        graph = wf.graph(detailed=True)
        dotfile = graph.create_dotfile_detailed(outdir=out_dir, name=name)
    else:
        raise Exception(
            f"type of the graph can be simple, detailed or nested, "
            f"but {plot_type} provided"
        )
    if not export:
        return dotfile
    else:
        if export is True:
            export = ["png"]
        elif isinstance(export, str):
            export = [export]
        formatted_dot = []
        for ext in export:
            formatted_dot.append(graph.export_graph(dotfile=dotfile, ext=ext))
        return dotfile, formatted_dot


def show_workflow(
    workflow_task: "workflow.Task",
    plot_type: str = "simple",
    lazy: ty.Sequence[str] | ty.Set[str] | None = None,
    use_lib: str | None = None,
    figsize: tuple[int, int] | None = None,
    **kwargs,
) -> None:
    """creating a graph and showing it"""
    out_dir = Path(tempfile.mkdtemp())
    png_graph = plot_workflow(
        workflow_task,
        out_dir=out_dir,
        plot_type=plot_type,
        export="png",
        lazy=lazy,
    )[1][0]

    if use_lib in ("matplotlib", None):
        try:
            import matplotlib.pyplot as plt
            from matplotlib.image import imread
        except ImportError:
            if use_lib == "matplotlib":
                raise ImportError(
                    "Please install either matplotlib to display the workflow image."
                )
        else:
            use_lib = "matplotlib"

            # Read the image
            img = imread(png_graph)

            if figsize is not None:
                plt.figure(figsize=figsize)
            # Display the image
            plt.imshow(img, **kwargs)
            plt.axis("off")
            plt.show()

    if use_lib in ("PIL", None):
        try:
            from PIL import Image
        except ImportError:
            msg = " or matplotlib" if use_lib is None else ""
            raise ImportError(
                f"Please install either Pillow{msg} to display the workflow image."
            )
        # Open the PNG image
        img = Image.open(png_graph)

        # Display the image
        img.show(**kwargs)


def attrs_fields(task, exclude_names=()) -> list[attrs.Attribute]:
    """Get the fields of a task, excluding some names."""
    return [field for field in task.__attrs_attrs__ if field.name not in exclude_names]


def attrs_values(obj, **kwargs) -> dict[str, ty.Any]:
    """Get the values of an attrs object."""
    return {
        n: v
        for n, v in attrs.asdict(obj, recurse=False, **kwargs).items()
        if not n.startswith("_")
    }


class _TaskFieldsList(dict[str, "Field"]):
    """A list of task fields. Acts like list in that you can iterate over the values
    but also access them like a dict or by attribute."""

    def __iter__(self):
        return iter(self.values())

    def __getattr__(self, name):
        return self[name]

    def __dir__(self):
        return sorted(self.keys())


def get_fields(task: "type[Task] | Task") -> _TaskFieldsList:
    """List the fields of a task"""
    if not inspect.isclass(task):
        task = type(task)
    if not attrs.has(task):
        return _TaskFieldsList()
    return _TaskFieldsList(
        **{
            f.name: f.metadata[PYDRA_ATTR_METADATA]
            for f in attrs.fields(task)
            if PYDRA_ATTR_METADATA in f.metadata
        }
    )


def asdict(obj, **kwargs) -> dict[str, ty.Any]:
    """Get the values of an attrs object."""
    return {f.name: getattr(obj, f.name) for f in get_fields(obj)}


def from_list_if_single(obj: ty.Any) -> ty.Any:
    """Converts a list to a single item if it is of length == 1"""
    from pydra.utils.typing import is_lazy

    if obj is attrs.NOTHING:
        return obj
    if is_lazy(obj):
        return obj
    if isinstance(obj, ty.Sequence) and not isinstance(obj, str):
        obj = list(obj)
        if len(obj) == 1:
            return obj[0]
    return obj


def wrap_text(text: str, width: int = 79, indent_size=4) -> str:
    """Wraps text to a given width, respecting word boundaries."""
    indent = " " * indent_size
    if len(text) <= width:
        return indent + text
    lines = []
    for line in text.splitlines():
        if len(line) > width:
            words = line.split()
            split_line = indent
            for word in words:
                if len(split_line) + len(word) + 1 > width:
                    lines.append(split_line.rstrip())
                    split_line = indent
                split_line += word + " "
            lines.append(split_line.rstrip())
        else:
            lines.append(indent + line)
    return "\n".join(lines).rstrip()


def task_help(
    task_type: "type[Task] | Task", line_width: int = 79, help_indent: int = 4
) -> list[str] | None:
    """Visit a job object and print its input/output interface."""
    import pydra.compose.base

    if isinstance(task_type, pydra.compose.base.Task):
        task_type = type(task_type)

    plugin_name = next(
        n
        for n, t in get_plugin_classes(pydra.compose, "Task").items()
        if issubclass(task_type, t)
    ).capitalize()

    header = f"Help for {plugin_name} task '{task_type.__name__}'"
    hyphen_line = "-" * len(header)
    lines = [hyphen_line, header, hyphen_line]
    inputs = get_fields(task_type)
    if inputs:
        lines.extend(["", "Inputs:"])
    if any(hasattr(i, "position") for i in inputs):
        inputs = sorted(inputs)
    for inpt in inputs:
        lines.extend(
            inpt.markdown_listing(
                line_width, help_indent=help_indent, as_input=True
            ).split("\n")
        )
    outputs = get_fields(task_type.Outputs)
    if outputs:
        lines.extend(["", "Outputs:"])
    for output in outputs:
        lines.extend(
            output.markdown_listing(line_width, help_indent=help_indent).split("\n")
        )
    lines.append("")
    return lines


def print_help(task: "Task[TaskType]") -> None:
    """Print help for a task."""
    lines = task_help(task)
    print("\n".join(lines))


def position_sort(args):
    """
    Sort objects by position, following Python indexing conventions.

    Ordering is positive positions, lowest to highest, followed by unspecified
    positions (``None``) and negative positions, lowest to highest.

    >>> position_sort([(None, "d"), (-3, "e"), (2, "b"), (-2, "f"), (5, "c"), (1, "a")])
    ['a', 'b', 'c', 'd', 'e', 'f']

    Parameters
    ----------
    args : list of (int/None, object) tuples

    Returns
    -------
    list of objects
    """
    import bisect

    pos, none, neg = [], [], []
    for entry in args:
        position = entry[0]
        if position is None:
            # Take existing order
            none.append(entry[1])
        elif position < 0:
            # Sort negatives while collecting
            bisect.insort(neg, entry)
        else:
            # Sort positives while collecting
            bisect.insort(pos, entry)

    return [arg for _, arg in pos] + none + [arg for _, arg in neg]


def ensure_list(obj, tuple2list=False):
    """
    Return a list whatever the input object is.

    Examples
    --------
    >>> ensure_list(list("abc"))
    ['a', 'b', 'c']
    >>> ensure_list("abc")
    ['abc']
    >>> ensure_list(tuple("abc"))
    [('a', 'b', 'c')]
    >>> ensure_list(tuple("abc"), tuple2list=True)
    ['a', 'b', 'c']
    >>> ensure_list(None)
    []
    >>> ensure_list(5.0)
    [5.0]

    """
    from pydra.utils.typing import is_lazy

    if obj is attrs.NOTHING:
        return attrs.NOTHING
    if obj is None:
        return []
    # list or numpy.array (this might need some extra flag in case an array has to be converted)
    elif isinstance(obj, list) or hasattr(obj, "__array__"):
        return obj
    elif tuple2list and isinstance(obj, tuple):
        return list(obj)
    elif is_lazy(obj):
        return obj
    # elif is_container(obj):
    #     raise NotImplementedError("just checking for now")
    return [obj]


def ensure_file_list(filename):
    """Return a list given either a string or a list."""
    if isinstance(filename, (str, bytes)):
        return [filename]
    elif isinstance(filename, list):
        return filename
    elif is_container(filename):
        return [x for x in filename]

    return None


# dj: copied from misc
def is_container(item):
    """
    Check if item is a container (list, tuple, dict, set).

    Parameters
    ----------
    item : :obj:`object`
        Input object to check.

    Returns
    -------
    output : :obj:`bool`
        ``True`` if container ``False`` otherwise.

    """
    if isinstance(item, str):
        return False
    elif hasattr(item, "__iter__"):
        return True

    return False


def is_workflow(obj):
    """Check whether an object is a :class:`Workflow` instance."""
    from pydra.compose.workflow import Task
    from pydra.engine.workflow import Workflow

    return isinstance(obj, (Task, Workflow))


def get_plugin_classes(namespace: types.ModuleType, class_name: str) -> dict[str, type]:
    """
    Get all classes within sub-packages of namespace package with a given name, e.g.
    "Worker" within "pydra.workers.*" sub-packages.

    Parameters
    ----------
    namespace : :obj:`str`
        The namespace to search for subclasses.
    base_class : :obj:`type`
        The base class to search for subclasses of.

    Returns
    -------
    :obj:`dict[str, type]`
        A dictionary mapping the sub-package name to classes of 'class_name' within
        the namespace package
    """
    sub_packages = [
        importlib.import_module(f"{namespace.__name__}.{m.name}")
        for m in pkgutil.iter_modules(namespace.__path__)
        if not m.name.startswith("base")
    ]
    return {
        pkg.__name__.split(".")[-1]: getattr(pkg, class_name)
        for pkg in sub_packages
        if hasattr(pkg, class_name)
    }


def unstructure(
    task_class: "type[Task]",
    filter: ty.Callable[[attrs.Attribute, ty.Any], bool] | None = None,
    value_serializer: (
        ty.Callable[[ty.Any, attrs.Attribute, ty.Any], ty.Any] | None
    ) = None,
    **kwargs: ty.Any,
) -> ty.Dict[str, ty.Any]:
    """Converts a Pydra task class into a dictionary representation that can be serialized
    (e.g. to a file), then reread and passed to an appropriate `pydra.compose.*.define`
    method to recreate the task.

    Parameters
    ----------
    task_class : type[pydra.compose.base.Task]
        The Pydra task class to convert.
    filter : callable, optional
        A function to filter out certain attributes from the task definition passed
        through to `attrs.asdict`. It should take an attribute and its value as
        arguments and return a boolean (True to keep the attribute, False to filter it out).
    value_serializer : callable, optional
        A function to serialize the value of an attribute. It should take the task
        definition, the attribute, and its value as arguments and return a serialized
        value.
    **kwargs : dict
        Additional keyword arguments to pass to `attrs.asdict` (i.e. all except `filter`,
        and `value_serializer`), e.g. `recurse`. See the `attrs` documentation for more
        details.

    Returns
    -------
    dict[str, ty.Any]
        A dictionary representation of the Pydra task.
    """
    from pydra.compose.base import Out

    if filter is None:
        filter = filter_out_defaults

    def full_val_serializer(
        obj: ty.Any, field: attrs.Attribute, value: ty.Any
    ) -> ty.Any:
        """A wrapper for the value serializer to handle the case where it is None."""
        if value_serializer is not None:
            value = value_serializer(obj, field, value)
        if isinstance(value, str):
            return value
        if isinstance(value, Mapping) and not isinstance(value, dict):
            # If the value is a mapping, convert it to a dict
            value = dict(value)
        elif isinstance(value, Collection) and not isinstance(value, list):
            # If the value is not a collection or string, convert it to a list
            value = list(value)
        return value

    input_fields = get_fields(task_class)
    executor = input_fields.pop(task_class._executor_name).default
    input_dicts = [
        attrs.asdict(i, filter=filter, value_serializer=full_val_serializer, **kwargs)
        for i in input_fields
        if (
            not isinstance(i, Out)  # filter out outarg fields
            and i.name not in task_class.BASE_ATTRS
        )
    ]
    output_dicts = [
        attrs.asdict(o, filter=filter, value_serializer=full_val_serializer, **kwargs)
        for o in get_fields(task_class.Outputs)
        if o.name not in task_class.Outputs.BASE_ATTRS
    ]
    dct = {
        "type": task_class._task_type(),
        task_class._executor_name: executor,
        "name": task_class.__name__,
        "inputs": {d.pop("name"): d for d in input_dicts},
        "outputs": {d.pop("name"): d for d in output_dicts},
    }
    class_attrs = {a: getattr(task_class, "_" + a) for a in task_class.TASK_CLASS_ATTRS}
    # Convert the frozensets to lists for serialization
    class_attrs["xor"] = [list(x) for x in class_attrs.pop("xor")]
    if value_serializer:
        # We need to create a mock attrs object for the class attrs to apply the
        # value_serializer to it
        mock_cls = _make_attrs_class(
            {a: task_class.__annotations__.get(a, ty.Any) for a in class_attrs}
        )
        attrs_fields = {f.name: f for f in attrs.fields(mock_cls)}
        mock = mock_cls(**class_attrs)
        class_attrs = {
            n: full_val_serializer(mock, attrs_fields[n], v)
            for n, v in class_attrs.items()
        }
    dct.update(class_attrs)

    return dct


def structure(task_class_dict: dict[str, ty.Any]) -> type["Task"]:
    """Converts a dictionary representation created by `unstructure` into a Pydra task
    class.

    Parameters
    ----------
    task_class_dict: dict[str, Any]
        the dictionary representation to unserialize

    Returns
    -------
    type[pydra.compose.base.Task]
        the unserialized task class
    """
    dct = copy(task_class_dict)
    task_type = dct.pop("type")
    mod = importlib.import_module(f"pydra.compose.{task_type}")
    return mod.define(dct.pop(mod.Task._executor_name), **dct)


def filter_out_defaults(atr: attrs.Attribute, value: ty.Any) -> bool:
    """Filter out values that match the attributes default value."""
    if isinstance(atr.default, attrs.Factory) and atr.default.factory() == value:
        return False
    if value == atr.default:
        return False
    return True


def _make_attrs_class(field_types: dict[str, type]) -> type:
    """Creates an attrs given the a dictionary of field names and their types.

    Parameters
    ----------
    field_types : dict[str, type]
        A dictionary mapping field names to their types.

    Returns
    -------
    type
        the attrs class.
    """
    return attrs.define(
        type(
            "MockAttrsClass",
            (),
            {n: attrs.field(type=t) for n, t in field_types.items()},
        )
    )



import typing as ty
import attrs
from attrs.converters import default_if_none


def donothing(*args: ty.Any, **kwargs: ty.Any) -> None:
    return None


@attrs.define(kw_only=True)
class TaskHooks:
    """Callable job hooks."""

    pre_run_task: ty.Callable = attrs.field(
        default=donothing, converter=default_if_none(donothing)
    )
    post_run_task: ty.Callable = attrs.field(
        default=donothing, converter=default_if_none(donothing)
    )
    pre_run: ty.Callable = attrs.field(
        default=donothing, converter=default_if_none(donothing)
    )
    post_run: ty.Callable = attrs.field(
        default=donothing, converter=default_if_none(donothing)
    )

    def reset(self):
        for val in ["pre_run_task", "post_run_task", "pre_run", "post_run"]:
            setattr(self, val, donothing)



"""Module to keep track of provenance information."""

import os
import typing as ty
import json
from pydra.utils.messenger import send_message, make_message, gen_uuid, now, AuditFlag
from pydra.utils.general import attrs_values
from fileformats.core import FileSet
from pydra.utils.hash import hash_function

try:
    import importlib_resources
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore

if ty.TYPE_CHECKING:
    from pydra.engine.job import Job


class Audit:
    """Handle provenance tracking and resource utilization."""

    def __init__(self, audit_flags, messengers, messenger_args, develop=None):
        """
        Initialize the auditing functionality.

        Parameters
        ----------
        audit_flags : :class:`AuditFlag`
            Base configuration of auditing.
        messengers : :obj:`pydra.util.messenger.Messenger`
            or list of :class:`pydra.util.messenger.Messenger`, optional
            Taskify types of messenger used by Audit to send a message.
            Could be `PrintMessenger`, `FileMessenger`, or `RemoteRESTMessenger`.
        messenger_args : :obj:`dict`, optional
            Optional arguments for the `Messenger.send` method.
        develop : :obj:`bool`, optional
            If True, the local context.jsonld file is used, otherwise the one from github is used.

        """
        from pydra.utils.general import ensure_list

        self.audit_flags = audit_flags
        self.messengers = ensure_list(messengers)
        self.messenger_args = messenger_args
        self.develop = develop

    def start_audit(self, odir):
        """
        Start recording provenance.

        Monitored information is not sent until directory is created,
        in case message directory is inside job output directory.

        Parameters
        ----------
        odir : :obj:`os.pathlike`
            Message output directory.

        """
        self.odir = odir
        if self.audit_check(AuditFlag.PROV):
            self.aid = f"uid:{gen_uuid()}"

            user_id = f"uid:{gen_uuid()}"
            start_message = {
                "@id": self.aid,
                "@type": "job",
                "startedAtTime": now(),
                "executedBy": user_id,
            }

        os.chdir(self.odir)
        if self.audit_check(AuditFlag.PROV):
            self.audit_message(start_message, AuditFlag.PROV)
        if self.audit_check(AuditFlag.RESOURCE):
            from pydra.utils.profiler import ResourceMonitor

            self.resource_monitor = ResourceMonitor(os.getpid(), logdir=self.odir)

    def monitor(self):
        """Start resource monitoring."""
        if self.audit_check(AuditFlag.RESOURCE):
            self.resource_monitor.start()
            if self.audit_check(AuditFlag.PROV):
                self.mid = f"uid:{gen_uuid()}"
                self.audit_message(
                    {
                        "@id": self.mid,
                        "@type": "monitor",
                        "startedAtTime": now(),
                        "wasStartedBy": self.aid,
                    },
                    AuditFlag.PROV,
                )

    def finalize_audit(self, result):
        """End auditing."""
        if self.audit_check(AuditFlag.RESOURCE):
            from pydra.engine.result import gather_runtime_info

            self.resource_monitor.stop()
            result.runtime = gather_runtime_info(self.resource_monitor.fname)
            if self.audit_check(AuditFlag.PROV):
                self.audit_message(
                    {"@id": self.mid, "endedAtTime": now(), "wasEndedBy": self.aid},
                    AuditFlag.PROV,
                )
                # audit resources/runtime information
                self.eid = f"uid:{gen_uuid()}"
                entity = attrs_values(result.runtime)
                entity.update(
                    **{
                        "@id": self.eid,
                        "@type": "runtime",
                        "prov:wasGeneratedBy": self.aid,
                    }
                )
                self.audit_message(entity, AuditFlag.PROV)
                self.audit_message(
                    {
                        "@type": "prov:Generation",
                        "entity_generated": self.eid,
                        "hadActivity": self.mid,
                    },
                    AuditFlag.PROV,
                )
            self.resource_monitor = None
        if self.audit_check(AuditFlag.PROV):
            # audit outputs
            self.audit_message(
                {"@id": self.aid, "endedAtTime": now(), "errored": result.errored},
                AuditFlag.PROV,
            )

    def audit_message(self, message, flags=None):
        """
        Send auditing message.

        Parameters
        ----------
        message : :obj:`dict`
            A message in Pydra is a JSON-LD message object.
        flags : :obj:`bool`, optional
            If True and self.audit_flag, the message is sent.

        """
        if self.develop:
            context_file = importlib_resources.files("pydra") / "schema/context.jsonld"
            with context_file.open() as fp:
                context = json.load(fp)
        else:
            context = {
                "@context": "https://raw.githubusercontent.com/nipype/pydra"
                "/master/pydra/schema/context.jsonld"
            }
        if self.audit_flags & flags:
            if self.messenger_args:
                send_message(
                    make_message(message, context=context),
                    messengers=self.messengers,
                    **self.messenger_args,
                )
            else:
                send_message(
                    make_message(message, context=context), messengers=self.messengers
                )

    def audit_check(self, flag):
        """
        Determine whether auditing is enabled for a particular flag.

        Parameters
        ----------
        flag : :obj: `bool`
            The flag that is checked.

        Returns
        -------
        bool
            Boolean AND for self.oudit_flags and flag
        """
        return self.audit_flags & flag

    def audit_task(self, job: "Job"):
        import subprocess as sp
        from pydra.utils.general import get_fields

        label = job.name

        command = job.task.cmdline if hasattr(job.task, "executable") else None
        attr_list = get_fields(job.task)
        for attrs in attr_list:
            input_name = attrs.name
            value = job.inputs[input_name]
            if isinstance(value, FileSet):
                input_path = os.path.abspath(value)
                file_hash = hash_function(value)
                entity_id = f"uid:{gen_uuid()}"
                entity_message = {
                    "@id": entity_id,
                    "Label": input_name,
                    "AtLocation": input_path,
                    "GeneratedBy": None,
                    "@type": "input",
                    "digest": file_hash,
                }
                self.audit_message(entity_message, AuditFlag.PROV)

        if command is not None:
            cmd_name = command.split()[0]
            software = f"{cmd_name} --version"
            # take the first word of command as the
            # name of the executable
            # (this may not always be the case)
            version_cmd = sp.run(software, shell=True, stdout=sp.PIPE).stdout.decode(
                "utf-8"
            )
            try:
                version_cmd = version_cmd.splitlines()[0]

            except IndexError:
                version_cmd = f"{cmd_name} -- Version unknown"

        else:
            version_cmd = None

        start_message = {
            "@id": self.aid,
            "@type": "job",
            "Label": label,
            "Command": command,
            "StartedAtTime": now(),
            "AssociatedWith": version_cmd,
        }

        self.audit_message(start_message, AuditFlag.PROV)



import os
import attrs
import typing as ty
import cloudpickle as cp
import concurrent.futures as cf
import logging
from pydra.engine.job import Job
from pydra.workers import base

logger = logging.getLogger("pydra.worker")

if ty.TYPE_CHECKING:
    from pydra.engine.result import Result


def get_available_cpus():
    """
    Return the number of CPUs available to the current process or, if that is not
    available, the total number of CPUs on the system.

    Returns
    -------
    n_proc : :obj:`int`
        The number of available CPUs.
    """
    # Will not work on some systems or if psutil is not installed.
    # See https://psutil.readthedocs.io/en/latest/#psutil.Process.cpu_affinity
    try:
        import psutil

        return len(psutil.Process().cpu_affinity())
    except (AttributeError, ImportError, NotImplementedError):
        pass

    # Not available on all systems, including macOS.
    # See https://docs.python.org/3/library/os.html#os.sched_getaffinity
    if hasattr(os, "sched_getaffinity"):
        return len(os.sched_getaffinity(0))

    # Last resort
    return os.cpu_count()


@attrs.define
class ConcurrentFuturesWorker(base.Worker):
    """A worker to execute in parallel using Python's concurrent futures."""

    n_procs: int = attrs.field(factory=get_available_cpus)
    pool: cf.ProcessPoolExecutor = attrs.field(
        eq=False, init=False, hash=False, repr=False
    )

    @pool.default
    def _pool_default(self) -> cf.ProcessPoolExecutor:
        return cf.ProcessPoolExecutor(self.n_procs)

    def __getstate__(self) -> dict[str, ty.Any]:
        """Return state for pickling."""
        state = super().__getstate__()
        del state["pool"]
        return state

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        """Set state from pickling."""
        super().__setstate__(state)
        self.pool = cf.ProcessPoolExecutor(self.n_procs)

    async def run(
        self,
        job: Job[base.TaskType],
        rerun: bool = False,
    ) -> "Result":
        """Run a job."""
        assert self.loop, "No event loop available to submit tasks"
        job_pkl = cp.dumps(job)
        return await self.loop.run_in_executor(
            self.pool, self.uncloudpickle_and_run, job_pkl, rerun
        )

    @classmethod
    def uncloudpickle_and_run(cls, job_pkl: bytes, rerun: bool) -> "Result":
        """Unpickle and run a job."""
        job: Job[base.TaskType] = cp.loads(job_pkl)
        return job.run(rerun=rerun)

    def close(self):
        """Finalize the internal pool of tasks."""
        self.pool.shutdown()


# Alias so it can be referred to as cf.Worker
Worker = ConcurrentFuturesWorker
